

ledger_pool_handle = None


def get_pool_handle():
    global ledger_pool_handle
    return ledger_pool_handle

def set_pool_handle(pool_handle):
    global ledger_pool_handle
    ledger_pool_handle = pool_handle

    


# Generated by Django 2.1.7 on 2019-04-09 00:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0004_auto_20190328_0248'),
    ]

    operations = [
        migrations.AlterField(
            model_name='agentconnection',
            name='partner_name',
            field=models.CharField(max_length=60),
        ),
        migrations.AlterField(
            model_name='agentconversation',
            name='connection_partner_name',
            field=models.CharField(max_length=60),
        ),
        migrations.AlterField(
            model_name='indycredentialdefinition',
            name='creddef_handle',
            field=models.CharField(max_length=80),
        ),
        migrations.AlterField(
            model_name='indycredentialdefinition',
            name='creddef_name',
            field=models.CharField(max_length=80),
        ),
        migrations.AlterField(
            model_name='indycredentialdefinition',
            name='ledger_creddef_id',
            field=models.CharField(max_length=80, unique=True),
        ),
        migrations.AlterField(
            model_name='indyorganization',
            name='org_name',
            field=models.CharField(max_length=60, unique=True),
        ),
        migrations.AlterField(
            model_name='indyproofrequest',
            name='proof_req_name',
            field=models.CharField(max_length=400, unique=True),
        ),
        migrations.AlterField(
            model_name='indyschema',
            name='ledger_schema_id',
            field=models.CharField(max_length=80, unique=True),
        ),
        migrations.AlterField(
            model_name='indyschema',
            name='schema_name',
            field=models.CharField(max_length=80),
        ),
        migrations.AlterField(
            model_name='indyschema',
            name='schema_version',
            field=models.CharField(max_length=80),
        ),
        migrations.AlterField(
            model_name='indysession',
            name='wallet_name',
            field=models.CharField(blank=True, max_length=60, null=True),
        ),
        migrations.AlterField(
            model_name='indyuser',
            name='first_name',
            field=models.CharField(blank=True, max_length=80),
        ),
        migrations.AlterField(
            model_name='indywallet',
            name='wallet_name',
            field=models.CharField(max_length=60, unique=True),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.views.generic.edit import UpdateView

from ..models import *


User = get_user_model()
TEST_USER_EMAIL = 'test@example.com'
TEST_USER_FIRST_NAME = 'Test'
TEST_USER_LAST_NAME = 'User'


class IndyWalletTests(TestCase):
    """
    Tests for Indy wallet model class
    """

    def test_wallet_create(self):
        # create a wallet
        my_wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        my_wallet.save()

        fetch_wallet = IndyWallet.objects.filter(wallet_name='test_wallet').all()
        self.assertEqual(len(fetch_wallet), 1)
        self.assertEqual(fetch_wallet[0].wallet_name, 'test_wallet')


class IndyUserTests(TestCase):
    """
    Tests for Indy custom User class.
    """

    def setUp(self):
        # Creates a single-user test database.
        self.user = User.objects.create(
            email=TEST_USER_EMAIL,
            first_name=TEST_USER_FIRST_NAME,
            last_name=TEST_USER_LAST_NAME,
        )

    def test_user_exists(self):
        # Tests user in setUp() does exists
        my_user = User.objects.filter(email=TEST_USER_EMAIL).all()[0]
        self.assertEqual(my_user.first_name, TEST_USER_FIRST_NAME)

    def test_user_with_wallet(self):
        # test we can create a user with a wallet
        my_wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        my_wallet.save()
        my_user = User.objects.filter(email=TEST_USER_EMAIL).all()[0]
        my_user.wallet = my_wallet
        my_user.save()

        fetch_user = User.objects.filter(email=TEST_USER_EMAIL).all()[0]
        self.assertEqual(my_user.wallet.wallet_name, my_wallet.wallet_name)


class IndyOrganizationTests(TestCase):
    """
    Tests for Indy organization class 
    """

    def test_organization_create(self):
        # tests creating an organization with a wallet 
        my_wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        my_wallet.save()
        my_org = IndyOrganization.objects.create(
            org_name='My Org',
            wallet=my_wallet,
        )
        my_org.save()

        fetch_org = IndyOrganization.objects.filter(org_name='My Org').all()
        self.assertEqual(len(fetch_org), 1)
        self.assertEqual(fetch_org[0].wallet.wallet_name, 'test_wallet')


class IndyOrgRelationshipTests(TestCase):
    """
    Tests for Indy organization/user relationship class 
    """

    def test_relationship_create(self):
        # tests creating a relationship between a user and organization
        user_wallet = IndyWallet.objects.create(
            wallet_name='user_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        user_wallet.save()
        my_user = User.objects.create(
            email='user@org.com',
            first_name='org',
            last_name='user',
            wallet=user_wallet,
        )
        org_wallet = IndyWallet.objects.create(
            wallet_name='org_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        org_wallet.save()
        my_org = IndyOrganization.objects.create(
            org_name='My Org',
            wallet=org_wallet,
        )
        my_org.save()
        my_relationship = IndyOrgRelationship.objects.create(
            org=my_org,
            user=my_user,
        )

        fetch_user = User.objects.filter(email='user@org.com').all()[0]
        self.assertEqual(len(fetch_user.indyrelationship_set.all()), 1)
        user_org = fetch_user.indyrelationship_set.all()[0].org
        self.assertEqual(user_org.org_name, 'My Org')

        fetch_org = IndyOrganization.objects.filter(org_name='My Org').all()[0]
        self.assertEqual(len(fetch_org.indyrelationship_set.all()), 1)
        org_user = fetch_org.indyrelationship_set.all()[0].user
        self.assertEqual(org_user.email, 'user@org.com')


class IndySchemaTests(TestCase):
    """
    Tests for IndySchema class
    """
    def test_schema_create(self):
        schema = IndySchema.objects.create(
            ledger_schema_id='123',
            schema_name='My Schema',
            schema_version='1.1.1',
            schema='this is the schema data',
            schema_template='template for adding credentials',
            schema_data='data written to the ledger',
        )
        schema.save()

        fetch_schema = IndySchema.objects.filter(ledger_schema_id='123').all()
        self.assertEqual(len(fetch_schema), 1)
        self.assertEqual(fetch_schema[0].schema_name, 'My Schema')


class IndyCredentialDefinitionTests(TestCase):
    """
    Tests for IndyCredentialDefinition class
    """
    def test_credentialdefinition_create(self):
        wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        wallet.save()
        schema = IndySchema.objects.create(
            ledger_schema_id='123',
            schema_name='My Schema',
            schema_version='1.1.1',
            schema='this is the schema data',
            schema_template='template for adding credentials',
            schema_data='data written to the ledger',
        )
        schema.save()
        cred_def = IndyCredentialDefinition.objects.create(
            ledger_creddef_id='456',
            ledger_schema=schema,
            wallet=wallet,
            creddef_name='my cred def',
            creddef_handle='4',
            creddef_template='a template for adding credentials',
            creddef_data='data written to the ledger',
        )
        cred_def.save()

        fetch_cred_def = IndyCredentialDefinition.objects.filter(ledger_creddef_id='456').all()
        self.assertEqual(len(fetch_cred_def), 1)
        self.assertEqual(fetch_cred_def[0].ledger_schema.schema_name, 'My Schema')
        self.assertEqual(fetch_cred_def[0].creddef_name, 'my cred def')


class IndyProofRequestTests(TestCase):
    """
    Tests for IndyProofRequest class
    """
    def test_proofrequest_create(self):
        proof_request = IndyProofRequest.objects.create(
            proof_req_name='test name',
            proof_req_description='a description',
            proof_req_attrs='revealed attributes',
            proof_req_predicates='zkp attributes',
        )
        proof_request.save()

        fetch_proof_request = IndyProofRequest.objects.filter(proof_req_name='test name').all()
        self.assertEqual(len(fetch_proof_request), 1)
        self.assertEqual(fetch_proof_request[0].proof_req_name, 'test name')


class AgentConnectionTests(TestCase):
    """
    Tests for Indy AgentConnection class
    """
    def test_connection_create(self):
        wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        wallet.save()
        connection = AgentConnection.objects.create(
            wallet=wallet,
            partner_name='partner',
            invitation='invitation to connect',
            token='token to identify connection',
            status='Active',
            connection_type='In or Out',
            connection_data='data representing connection state',
        )
        connection.save()

        fetch_connection = AgentConnection.objects.filter(wallet=wallet, partner_name='partner').all()
        self.assertEqual(len(fetch_connection), 1)
        self.assertEqual(fetch_connection[0].partner_name, 'partner')


class AgentConversationTests(TestCase):
    """
    Tests for Indy AgentConversation class
    """
    def test_conversation_create(self):
        wallet = IndyWallet.objects.create(
            wallet_name='test_wallet',
            wallet_config='{"some":"test", "string":"."}',
        )
        wallet.save()
        connection = AgentConnection.objects.create(
            wallet=wallet,
            partner_name='partner',
            invitation='invitation to connect',
            token='token to identify connection',
            status='Active',
            connection_type='In or Out',
            connection_data='data representing connection state',
        )
        connection.save()
        conversation = AgentConversation.objects.create(
            connection=connection,
            conversation_type='proof or credential',
            message_id='123',
            status='Active',
            proof_state='only for proofs',
            conversation_data='data representing conversation state',
        )
        conversation.save()

        fetch_conversation = AgentConversation.objects.filter(connection__wallet=wallet, connection=connection, message_id='123').all()
        self.assertEqual(len(fetch_conversation), 1)
        self.assertEqual(fetch_conversation[0].conversation_data, 'data representing conversation state')




# Generated by Django 2.1.7 on 2019-03-27 13:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0002_auto_20190319_2207'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndyOrgRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=40, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name='indyorganization',
            name='role',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyOrgRole'),
        ),
        migrations.AddField(
            model_name='indyschema',
            name='roles',
            field=models.ManyToManyField(to='indy_community.IndyOrgRole'),
        ),
    ]



from django import forms
from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth import get_user_model
from django.db.models import Q

import json

from .models import *


###############################################################
# Forms to request a connection to a mobile wallet (user only)
###############################################################
class RequestMobileConnectionForm(UserCreationForm):
    email = forms.CharField(label='Email', max_length=120)
    org = forms.ModelChoiceField(label='Organization', queryset=IndyOrganization.objects.filter(Q(role__name='School') | Q(role__name='Employer') | Q(role__name='Bank') | Q(role__name='Goverment')).all())

    class Meta:
        model = get_user_model()
        fields = ('first_name', 'last_name', 'email', 'password1', 'password2', 'org')


###############################################################
# Forms to support user and organization registration
###############################################################
class UserSignUpForm(UserCreationForm):
    first_name = forms.CharField(max_length=80, label='First Name', required=False,
                                 help_text='Optional.')
    last_name = forms.CharField(max_length=150, label='Last Name', required=False,
                                 help_text='Optional.')
    email = forms.EmailField(max_length=254, label='Email Address', required=True,
                                 help_text='Required. Provide a valid email address.')

    class Meta:
        model = get_user_model()
        fields = ('first_name', 'last_name', 'email', 'password1', 'password2')


class OrganizationSignUpForm(UserSignUpForm):
    org_name = forms.CharField(max_length=60, label='Company Name', required=True,
                                 help_text='Required.')
    org_role_name = forms.CharField(max_length=40, label='Company Role', required=True,
                                 help_text='Required.')
    ico_url = forms.CharField(max_length=120, label="URL for company logo", required=False)


######################################################################
# forms to create and confirm agent-to-agent connections
######################################################################
class WalletNameForm(forms.Form):
    wallet_name = forms.CharField(widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs):
        super(WalletNameForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True

class VisibleWalletNameForm(forms.Form):
    wallet_name = forms.CharField(max_length=60)

    def __init__(self, *args, **kwargs):
        super(VisibleWalletNameForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True


class SendConnectionInvitationForm(WalletNameForm):
    partner_name = forms.CharField(label='Partner Name', max_length=60)

    def __init__(self, *args, **kwargs):
        super(SendConnectionInvitationForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True


class SendConnectionResponseForm(SendConnectionInvitationForm):
    connection_id = forms.IntegerField(widget=forms.HiddenInput())
    invitation_details = forms.CharField(label='Invitation', max_length=4000, widget=forms.Textarea)

    def __init__(self, *args, **kwargs):
        super(SendConnectionResponseForm, self).__init__(*args, **kwargs)
        self.fields['connection_id'].widget.attrs['readonly'] = True


class PollConnectionStatusForm(VisibleWalletNameForm):
    connection_id = forms.IntegerField(label="Id")

    def __init__(self, *args, **kwargs):
        super(PollConnectionStatusForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['connection_id'].widget.attrs['readonly'] = True


######################################################################
# forms to offer, request, send and receive credentials
######################################################################
class SendConversationResponseForm(WalletNameForm):
    conversation_id = forms.IntegerField(widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs):
        super(SendConversationResponseForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True
        self.fields['conversation_id'].widget.attrs['readonly'] = True


class PollConversationStatusForm(VisibleWalletNameForm):
    conversation_id = forms.IntegerField(label="Id")

    def __init__(self, *args, **kwargs):
        super(PollConversationStatusForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['conversation_id'].widget.attrs['readonly'] = True


class SelectCredentialOfferForm(WalletNameForm):
    connection_id = forms.IntegerField(widget=forms.HiddenInput())
    partner_name = forms.CharField(label='Partner Name', max_length=60)
    cred_def = forms.ModelChoiceField(label='Cred Def', queryset=IndyCredentialDefinition.objects.all())

    def __init__(self, *args, **kwargs):
        super(SelectCredentialOfferForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True
        self.fields['connection_id'].widget.attrs['readonly'] = True
        self.fields['partner_name'].widget.attrs['readonly'] = True

        # build a list of Credential Definitions available to the current wallet
        initial = kwargs.get('initial')
        if initial:
            wallet_name = initial.get('wallet_name')
            self.fields['cred_def'].queryset = IndyCredentialDefinition.objects.filter(wallet__wallet_name=wallet_name).all()


class SendCredentialOfferForm(WalletNameForm):
    connection_id = forms.IntegerField(widget=forms.HiddenInput())
    partner_name = forms.CharField(label='Partner Name', max_length=60)
    cred_def = forms.CharField(max_length=80, widget=forms.HiddenInput())
    credential_name = forms.CharField(label='Credential Name', max_length=80)
    credential_tag = forms.CharField(max_length=80, widget=forms.HiddenInput())
    schema_attrs = forms.CharField(label='Credential Attributes', max_length=4000, widget=forms.Textarea)

    def __init__(self, *args, **kwargs):
        super(SendCredentialOfferForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True
        self.fields['connection_id'].widget.attrs['readonly'] = True
        self.fields['partner_name'].widget.attrs['readonly'] = True
        self.fields['cred_def'].widget.attrs['readonly'] = True

        # build a list of attributes for the given schema
        initial = kwargs.get('initial')
        if initial:
            schema_attrs = initial.get('schema_attrs', '{}')
            schema_attrs = json.loads(schema_attrs)
            self.fields['schema_attrs'].widget.attrs['hidden'] = True
            for attr in schema_attrs:
                field_name = 'schema_attr_' + attr
                self.fields[field_name] = forms.CharField(label=attr, max_length=200)


class SendCredentialResponseForm(SendConversationResponseForm):
    # a bunch of fields that are read-only to present to the user
    from_partner_name = forms.CharField(label='Partner Name', max_length=60)
    claim_id = forms.CharField(max_length=80, widget=forms.HiddenInput())
    claim_name = forms.CharField(label='Credential Name', max_length=400)
    libindy_offer_schema_id = forms.CharField(max_length=120, widget=forms.HiddenInput())
    credential_attrs = forms.CharField(label='Credential Attrs', max_length=4000, widget=forms.Textarea)

    def __init__(self, *args, **kwargs):
        super(SendCredentialResponseForm, self).__init__(*args, **kwargs)
        self.fields['from_partner_name'].widget.attrs['readonly'] = True
        self.fields['claim_id'].widget.attrs['readonly'] = True
        self.fields['claim_name'].widget.attrs['readonly'] = True
        self.fields['libindy_offer_schema_id'].widget.attrs['readonly'] = True
        self.fields['credential_attrs'].widget.attrs['readonly'] = True

        # build a list of attributes for the current schema
        initial = kwargs.get('initial')
        if initial:
            credential_attrs = initial.get('credential_attrs', {})
            self.fields['credential_attrs'].widget.attrs['hidden'] = True
            for attr in credential_attrs:
                field_name = 'credential_attr_' + attr
                self.fields[field_name] = forms.CharField(label=attr, initial=credential_attrs[attr])
                self.fields[field_name].widget.attrs['readonly'] = True


######################################################################
# forms to request, send and receive proofs
######################################################################
class SelectProofRequestForm(WalletNameForm):
    connection_id = forms.IntegerField(widget=forms.HiddenInput())
    partner_name = forms.CharField(label='Partner Name', max_length=60)
    proof_request = forms.ModelChoiceField(label='Proof Request Type', queryset=IndyProofRequest.objects.all())

    def __init__(self, *args, **kwargs):
        super(SelectProofRequestForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True
        self.fields['connection_id'].widget.attrs['readonly'] = True
        self.fields['partner_name'].widget.attrs['readonly'] = True


class SendProofRequestForm(WalletNameForm):
    connection_id = forms.IntegerField(widget=forms.HiddenInput())
    partner_name = forms.CharField(label='Partner Name', max_length=60)
    proof_name = forms.CharField(label='Proof Name', max_length=400)
    proof_attrs = forms.CharField(label='Proof Attributes', max_length=4000, widget=forms.Textarea)
    proof_predicates = forms.CharField(label='Proof Predicates', max_length=4000, widget=forms.Textarea)

    def __init__(self, *args, **kwargs):
        super(SendProofRequestForm, self).__init__(*args, **kwargs)
        self.fields['wallet_name'].widget.attrs['readonly'] = True
        self.fields['wallet_name'].widget.attrs['hidden'] = True
        self.fields['connection_id'].widget.attrs['readonly'] = True
        self.fields['partner_name'].widget.attrs['readonly'] = True


class SendProofReqResponseForm(SendConversationResponseForm):
    # a bunch of fields that are read-only to present to the user
    from_partner_name = forms.CharField(label='Partner Name', max_length=60)
    proof_req_name = forms.CharField(label='Proof Request Name', max_length=400)

    def __init__(self, *args, **kwargs):
        super(SendProofReqResponseForm, self).__init__(*args, **kwargs)
        self.fields['from_partner_name'].widget.attrs['readonly'] = True
        self.fields['proof_req_name'].widget.attrs['readonly'] = True


class SelectProofReqClaimsForm(SendProofReqResponseForm):
    requested_attrs = forms.CharField(label='Requested Attrs', widget=forms.HiddenInput)

    def __init__(self, *args, **kwargs):
        super(SelectProofReqClaimsForm, self).__init__(*args, **kwargs)

        # list requested attributes and the available claims, for the user to select
        initial = kwargs.get('initial')
        if initial:
            field_attrs = initial.get('requested_attrs', '{}')
            if 'attrs' in field_attrs:
                for attr in field_attrs['attrs']:
                    field_name = 'proof_req_attr_' + attr
                    choices = []
                    claim_no = 0
                    if 0 < len(field_attrs['attrs'][attr]):
                        for claim in field_attrs['attrs'][attr]:
                            choices.append(('ref::'+claim['cred_info']['referent'], json.dumps(claim['cred_info']['attrs'])))
                            claim_no = claim_no + 1
                        self.fields[field_name] = forms.ChoiceField(label='Select claim for '+attr, choices=tuple(choices), widget=forms.RadioSelect())
                    else:
                        self.fields[field_name] = forms.CharField(label='No claims available for '+attr+', enter value:', max_length=80)




from .model_tests import IndyUserTests, IndyWalletTests, IndyOrganizationTests, IndyOrgRelationshipTests, IndySchemaTests, IndyCredentialDefinitionTests, IndyProofRequestTests, AgentConnectionTests, AgentConversationTests
from .wallet_util_tests import WalletDBTests
from .indy_util_tests import IndyDIDTests
from .registration_util_tests import RegistrationTests
from .agent_util_tests import AgentInteractionTests



# Generated by Django 2.1.7 on 2019-03-16 00:40

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0009_alter_user_last_name_max_length'),
        ('sessions', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndyUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('email', models.EmailField(max_length=254, unique=True)),
                ('first_name', models.CharField(blank=True, max_length=30)),
                ('last_name', models.CharField(blank=True, max_length=150)),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now)),
                ('last_login', models.DateTimeField(blank=True, null=True)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AgentConnection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('partner_name', models.CharField(max_length=30)),
                ('invitation', models.TextField(blank=True, max_length=4000)),
                ('token', models.CharField(blank=True, max_length=80)),
                ('status', models.CharField(max_length=20)),
                ('connection_type', models.CharField(max_length=20)),
                ('connection_data', models.TextField(blank=True, max_length=4000)),
            ],
        ),
        migrations.CreateModel(
            name='AgentConversation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('connection_partner_name', models.CharField(max_length=30)),
                ('conversation_type', models.CharField(max_length=30)),
                ('message_id', models.CharField(max_length=30)),
                ('status', models.CharField(max_length=20)),
                ('proof_state', models.CharField(blank=True, max_length=20)),
                ('conversation_data', models.TextField(blank=True, max_length=4000)),
            ],
        ),
        migrations.CreateModel(
            name='IndyCredentialDefinition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ledger_creddef_id', models.CharField(max_length=40, unique=True)),
                ('creddef_name', models.CharField(max_length=40)),
                ('creddef_handle', models.CharField(max_length=40)),
                ('creddef_template', models.TextField(max_length=4000)),
                ('creddef_data', models.TextField(max_length=4000)),
            ],
        ),
        migrations.CreateModel(
            name='IndyOrganization',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('org_name', models.CharField(max_length=40, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='IndyOrgRelationship',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('org', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyOrganization')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyUser')),
            ],
        ),
        migrations.CreateModel(
            name='IndyProofRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('proof_req_name', models.CharField(max_length=40, unique=True)),
                ('proof_req_description', models.TextField(max_length=4000)),
                ('proof_req_attrs', models.TextField(max_length=4000)),
                ('proof_req_predicates', models.TextField(blank=True, max_length=4000)),
            ],
        ),
        migrations.CreateModel(
            name='IndySchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ledger_schema_id', models.CharField(max_length=40, unique=True)),
                ('schema_name', models.CharField(max_length=40)),
                ('schema_version', models.CharField(max_length=40)),
                ('schema', models.TextField(max_length=4000)),
                ('schema_template', models.TextField(max_length=4000)),
                ('schema_data', models.TextField(max_length=4000)),
            ],
        ),
        migrations.CreateModel(
            name='IndySession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('wallet_name', models.CharField(blank=True, max_length=30, null=True)),
                ('session', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='sessions.Session')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyUser')),
            ],
        ),
        migrations.CreateModel(
            name='IndyWallet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('wallet_name', models.CharField(max_length=30, unique=True)),
                ('wallet_config', models.TextField(blank=True, max_length=4000)),
            ],
        ),
        migrations.AddField(
            model_name='indyorganization',
            name='wallet',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
        migrations.AddField(
            model_name='indycredentialdefinition',
            name='ledger_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndySchema'),
        ),
        migrations.AddField(
            model_name='indycredentialdefinition',
            name='wallet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
        migrations.AddField(
            model_name='agentconversation',
            name='wallet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
        migrations.AddField(
            model_name='agentconnection',
            name='wallet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
        migrations.AddField(
            model_name='indyuser',
            name='wallet',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
    ]



import asyncio
import aiohttp
import json
from os import environ
import random

from django.conf import settings

from indy import anoncreds, crypto, did, ledger, pool, wallet
from indy.error import ErrorCode, IndyError

from .models import *
from .utils import *


def calc_wallet_seed(wallet_name, org_role=''):
    """
    calculate a wallet seed based on the wallet wallet name
    """

    if org_role == 'Trustee':
        return settings.INDY_CONFIG['vcx_enterprise_seed']
    else:
        return (settings.INDY_CONFIG['vcx_institution_seed'] + wallet_name)[-32:]


async def register_did_on_ledger(ledger_url, alias, seed):
    """
    Register DID on ledger via VON Network ledger browser
    """

    try:
        async with aiohttp.ClientSession() as client:
            response = await client.post(
                "{}/register".format(ledger_url),
                json={"alias": alias, "seed": seed, "role": "TRUST_ANCHOR"},
            )
            nym_info = await response.json()
            print(nym_info)
    except Exception as e:
        raise Exception(str(e)) from None
    if not nym_info or not nym_info["did"]:
        raise Exception(
            "DID registration failed: {}".format(nym_info)
        )
    return nym_info


def create_and_register_did(wallet_name, org_role):
    """
    Register DID on ledger (except Trustee which is assumed already there)
    """

    if org_role == 'Trustee':
        # don't register Trustee role
        return

    if not settings.INDY_CONFIG['register_dids']:
        return

    enterprise_seed = calc_wallet_seed(wallet_name, org_role)
    ledger_url = settings.INDY_CONFIG['ledger_url']
    nym_info = run_coroutine_with_args(register_did_on_ledger, ledger_url, wallet_name, enterprise_seed)

    return nym_info



# Generated by Django 2.1.7 on 2019-04-13 01:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0006_auto_20190413_0001'),
    ]

    operations = [
        migrations.AlterField(
            model_name='agentconnection',
            name='partner_name',
            field=models.CharField(max_length=60),
        ),
        migrations.AlterField(
            model_name='agentconversation',
            name='connection',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='indy_community.AgentConnection'),
        ),
    ]



"""Indy implementation of BaseStorage interface."""

import json
from typing import Mapping, Sequence

from indy import non_secrets
from indy.error import IndyError, ErrorCode

from .base import BaseStorage, BaseStorageRecordSearch
from .error import (
    StorageError,
    StorageDuplicateError,
    StorageNotFoundError,
    StorageSearchError,
)
from .record import StorageRecord
#from ..wallet.indy import IndyWallet


def _validate_record(record: StorageRecord):
    if not record:
        raise StorageError("No record provided")
    if not record.id:
        raise StorageError("Record has no ID")
    if not record.type:
        raise StorageError("Record has no type")


class IndyStorage(BaseStorage):
    """Indy Non-Secrets interface."""

    def __init__(self, wallet_handle):
        """
        Initialize a `BasicStorage` instance.

        Args:
            wallet: The indy wallet instance to use

        """
        self._wallet_handle = wallet_handle

    @property
    def wallet_handle(self):
        """Accessor for wallet handle."""
        return self._wallet_handle

    async def add_record(self, record: StorageRecord):
        """
        Add a new record to the store.

        Args:
            record: `StorageRecord` to be stored

        """
        _validate_record(record)
        tags_json = json.dumps(record.tags) if record.tags else None
        try:
            await non_secrets.add_wallet_record(
                self._wallet_handle, record.type, record.id, record.value, tags_json
            )
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemAlreadyExists:
                raise StorageDuplicateError("Duplicate record ID: {}".format(record.id))
            raise StorageError(str(x_indy))

    async def get_record(self, record_type: str, record_id: str) -> StorageRecord:
        """
        Fetch a record from the store by type and ID.

        Args:
            record_type: The record type
            record_id: The record id

        Returns:
            A `StorageRecord` instance

        Raises:
            StorageError: If the record is not provided
            StorageError: If the record ID not provided
            StorageNotFoundError: If the record is not found
            StorageError: If record not found

        """
        if not record_type:
            raise StorageError("Record type not provided")
        if not record_id:
            raise StorageError("Record ID not provided")
        options_json = json.dumps(
            {"retrieveType": True, "retrieveValue": True, "retrieveTags": True}
        )
        try:
            result_json = await non_secrets.get_wallet_record(
                self._wallet_handle, record_type, record_id, options_json
            )
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                raise StorageNotFoundError("Record not found: {}".format(record_id))
            raise StorageError(str(x_indy))
        result = json.loads(result_json)
        return StorageRecord(
            type=result["type"],
            id=result["id"],
            value=result["value"],
            tags=result["tags"] or {},
        )

    async def update_record_value(self, record: StorageRecord, value: str):
        """
        Update an existing stored record's value.

        Args:
            record: `StorageRecord` to update
            value: The new value

        Raises:
            StorageNotFoundError: If record not found
            StorageError: If a libindy error occurs

        """
        _validate_record(record)
        try:
            await non_secrets.update_wallet_record_value(
                self._wallet_handle, record.type, record.id, value
            )
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                raise StorageNotFoundError("Record not found: {}".format(record.id))
            raise StorageError(str(x_indy))

    async def update_record_tags(self, record: StorageRecord, tags: Mapping):
        """
        Update an existing stored record's tags.

        Args:
            record: `StorageRecord` to update
            tags: New tags

        Raises:
            StorageNotFoundError: If record not found
            StorageError: If a libindy error occurs

        """
        _validate_record(record)
        tags_json = json.dumps(tags) if tags else "{}"
        try:
            await non_secrets.update_wallet_record_tags(
                self._wallet_handle, record.type, record.id, tags_json
            )
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                raise StorageNotFoundError("Record not found: {}".format(record.id))
            raise StorageError(str(x_indy))

    async def delete_record_tags(
        self, record: StorageRecord, tags: (Sequence, Mapping)
    ):
        """
        Update an existing stored record's tags.

        Args:
            record: `StorageRecord` to delete
            tags: Tags

        """
        _validate_record(record)
        if tags:
            # check existence of record first (otherwise no exception thrown)
            await self.get_record(record.type, record.id)

            tag_names_json = json.dumps(list(tags))
            await non_secrets.delete_wallet_record_tags(
                self._wallet_handle, record.type, record.id, tag_names_json
            )

    async def delete_record(self, record: StorageRecord):
        """
        Delete a record.

        Args:
            record: `StorageRecord` to delete

        Raises:
            StorageNotFoundError: If record not found
            StorageError: If a libindy error occurs

        """
        _validate_record(record)
        try:
            await non_secrets.delete_wallet_record(
                self._wallet_handle, record.type, record.id
            )
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                raise StorageNotFoundError("Record not found: {}".format(record.id))
            raise StorageError(str(x_indy))

    def search_records(
        self, type_filter: str, tag_query: Mapping = None, page_size: int = None
    ) -> "IndyStorageRecordSearch":
        """
        Search stored records.

        Args:
            type_filter: Filter string
            tag_query: Tags to query
            page_size: Page size

        Returns:
            An instance of `BaseStorageRecordSearch`

        """
        return IndyStorageRecordSearch(self, type_filter, tag_query, page_size)


class IndyStorageRecordSearch(BaseStorageRecordSearch):
    """Represent an active stored records search."""

    def __init__(
        self,
        store: IndyStorage,
        type_filter: str,
        tag_query: Mapping,
        page_size: int = None,
    ):
        """
        Initialize a `IndyStorageRecordSearch` instance.

        Args:
            store: `BaseStorage` to search
            type_filter: Filter string
            tag_query: Tags to search
            page_size: Size of page to return

        """
        super(IndyStorageRecordSearch, self).__init__(
            store, type_filter, tag_query, page_size
        )
        self._handle = None

    @property
    def opened(self) -> bool:
        """
        Accessor for open state.

        Returns:
            True if opened, else False

        """
        return self._handle is not None

    @property
    def handle(self):
        """
        Accessor for search handle.

        Returns:
            The handle

        """
        return self._handle

    async def fetch(self, max_count: int) -> Sequence[StorageRecord]:
        """
        Fetch the next list of results from the store.

        Args:
            max_count: Max number of records to return

        Returns:
            A list of `StorageRecord`s

        Raises:
            StorageSearchError: If the search query has not been opened

        """
        if not self.opened:
            raise StorageSearchError("Search query has not been opened")
        result_json = await non_secrets.fetch_wallet_search_next_records(
            self.store.wallet_handle, self._handle, max_count
        )
        results = json.loads(result_json)
        ret = []
        if results["records"]:
            for row in results["records"]:
                ret.append(
                    StorageRecord(
                        type=row["type"],
                        id=row["id"],
                        value=row["value"],
                        tags=row["tags"],
                    )
                )
        return ret

    async def open(self):
        """Start the search query."""
        query_json = json.dumps(self.tag_query or {})
        options_json = json.dumps(
            {
                "retrieveRecords": True,
                "retrieveTotalCount": True,
                "retrieveType": True,
                "retrieveValue": True,
                "retrieveTags": True,
            }
        )
        self._handle = await non_secrets.open_wallet_search(
            self.store.wallet_handle, self.type_filter, query_json, options_json
        )

    async def close(self):
        """Dispose of the search query."""
        if self._handle:
            await non_secrets.close_wallet_search(self._handle)
            self._handle = None



import asyncio
import aiohttp
import json
from os import environ
from pathlib import Path
from tempfile import gettempdir
import random
import uuid

from django.conf import settings

from indy.error import ErrorCode, IndyError

from vcx.api.connection import Connection
from vcx.api.schema import Schema
from vcx.api.credential_def import CredentialDef
from vcx.api.credential import Credential
from vcx.state import State, ProofState
from vcx.api.disclosed_proof import DisclosedProof
from vcx.api.issuer_credential import IssuerCredential
from vcx.api.proof import Proof
from vcx.api.utils import vcx_agent_provision
from vcx.api.vcx_init import vcx_init_with_config
from vcx.common import shutdown

from .models import *
from .utils import *



DUMMY_SEED = "00000000000000000000000000000000"

######################################################################
# utilities to provision vcx agents
######################################################################
def vcx_provision_config(wallet_name, raw_password, institution_name, did_seed=None, org_role='', institution_logo_url='http://robohash.org/456'):
    """
    Build a configuration object for a VCX environment or agent
    """

    provisionConfig = {
        'agency_url': settings.INDY_CONFIG['vcx_agency_url'],
        'agency_did': settings.INDY_CONFIG['vcx_agency_did'],
        'agency_verkey': settings.INDY_CONFIG['vcx_agency_verkey'],
        'pool_name': 'pool_' + wallet_name,
        'wallet_type': 'postgres_storage',
        'wallet_name': wallet_name,
        'wallet_key': raw_password,
        'storage_config': json.dumps(settings.INDY_CONFIG['storage_config']),
        'storage_credentials': json.dumps(settings.INDY_CONFIG['storage_credentials']),
        'payment_method': settings.INDY_CONFIG['vcx_payment_method'],
    }

    # role-dependant did seed
    if did_seed and 0 < len(did_seed):
        provisionConfig['enterprise_seed'] = did_seed
    else:
        provisionConfig['enterprise_seed'] = DUMMY_SEED

    provisionConfig['institution_name'] = institution_name
    provisionConfig['institution_logo_url'] = institution_logo_url
    provisionConfig['genesis_path'] = settings.INDY_CONFIG['vcx_genesis_path']
    provisionConfig['pool_name'] = 'pool_' + wallet_name

    return provisionConfig


def initialize_and_provision_vcx(wallet_name, raw_password, institution_name, did_seed=None, org_role='', institution_logo_url='http://robohash.org/456'):
    """
    Provision a wallet as a VCX Agent.
    """

    provisionConfig = vcx_provision_config(wallet_name, raw_password, institution_name, did_seed=did_seed, org_role=org_role, institution_logo_url=institution_logo_url)

    print(" >>> Provision an agent and wallet, get back configuration details")
    try:
        provisionConfig_json = json.dumps(provisionConfig)
        config = run_coroutine_with_args(vcx_agent_provision, provisionConfig_json)
    except:
        raise

    config = json.loads(config)

    # Set some additional configuration options specific to alice
    config['institution_name'] = institution_name
    config['institution_logo_url'] = institution_logo_url
    config['genesis_path'] = settings.INDY_CONFIG['vcx_genesis_path']
    config['pool_name'] = 'pool_' + wallet_name

    try:
        config_json = json.dumps(config)
        run_coroutine_with_args(vcx_init_with_config, config_json)
    except:
        raise
    finally:
        try:
            shutdown(False)
        except:
            raise

    return json.dumps(config)


######################################################################
# utilities to create schemas and credential defitions
######################################################################

def create_schema_json(schema_name, schema_version, schema_attrs):
    """
    Create an Indy Schema object based on a list of attributes.
    Returns the schema as well as a template for creating credentials.
    """

    schema = {
        'name': schema_name,
        'version': schema_version,
        'attributes': schema_attrs
    }
    creddef_template = {}
    for attr in schema_attrs:
        creddef_template[attr] = ''

    return (json.dumps(schema), json.dumps(creddef_template))


def create_schema(wallet, schema_json, schema_template, initialize_vcx=True):
    """
    Create an Indy Schema (VCX) and also store in our local database.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    try:
        schema = json.loads(schema_json)
        vcxschema = run_coroutine_with_args(Schema.create, 'schema_uuid', schema['name'], schema['version'], schema['attributes'], 0)
        schema_id = run_coroutine(vcxschema.get_schema_id)
        schema_data = run_coroutine(vcxschema.serialize)

        indy_schema = IndySchema(
                            ledger_schema_id = schema_id,
                            schema_name = schema['name'],
                            schema_version = schema['version'],
                            schema = schema_data,
                            schema_template = schema_template,
                            schema_data = json.dumps(schema_data)
                            )
        indy_schema.save()

    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return indy_schema


def create_creddef(wallet, indy_schema, creddef_name, creddef_template, initialize_vcx=True):
    """
    Create an Indy Credential Definition (VCX) and also store in our local database
    """

    # wallet specific-configuration for creatig the cred def
    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    try:
        cred_def = run_coroutine_with_args(CredentialDef.create, 'credef_uuid', creddef_name, indy_schema.ledger_schema_id, 0)
        cred_def_handle = cred_def.handle
        cred_def_id = run_coroutine(cred_def.get_cred_def_id)
        creddef_data = run_coroutine(cred_def.serialize)

        indy_creddef = IndyCredentialDefinition(
                            ledger_creddef_id = cred_def_id,
                            ledger_schema = indy_schema,
                            wallet = wallet,
                            creddef_name = creddef_name,
                            creddef_handle = cred_def_handle,
                            creddef_template = creddef_template,
                            creddef_data = json.dumps(creddef_data)
                            )
        indy_creddef.save()

    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return indy_creddef


def create_proof_request(name, description, attrs, predicates):
    """
    Create a proof request template (local database only).
    """

    proof_req_attrs = json.dumps(attrs)
    proof_req_predicates = json.dumps(predicates)
    proof_request = IndyProofRequest(
                            proof_req_name = name,
                            proof_req_description = description,
                            proof_req_attrs = proof_req_attrs,
                            proof_req_predicates = proof_req_predicates
                            )
    proof_request.save()

    return proof_request


######################################################################
# utilities to create and confirm agent-to-agent connections
######################################################################

def send_connection_invitation(wallet, partner_name, initialize_vcx=True):
    """
    Create a VCX Connection Invitation.
    Creates a record for the initator only (receiver is checked in the corresponding view).
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        connection_to_ = run_coroutine_with_args(Connection.create, partner_name)
        run_coroutine_with_args(connection_to_.connect, '{"use_public_did": true}')
        run_coroutine(connection_to_.update_state)
        invite_details = run_coroutine_with_args(connection_to_.invite_details, False)

        connection_data = run_coroutine(connection_to_.serialize)
        connection_to_.release()
        connection_to_ = None

        connection = AgentConnection(
            wallet = wallet,
            partner_name = partner_name,
            invitation = json.dumps(invite_details),
            token = str(uuid.uuid4()),
            connection_type = 'Outbound',
            connection_data = json.dumps(connection_data),
            status = 'Sent')
        connection.save()
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return connection


def send_connection_confirmation(wallet, connection_id, partner_name, invite_details, initialize_vcx=True):
    """
    Send a confirmation message for a VCX Invitation.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        connection_from_ = run_coroutine_with_args(Connection.create_with_details, partner_name, invite_details)
        connection_data = run_coroutine(connection_from_.serialize)
        run_coroutine_with_args(connection_from_.connect, '{"use_public_did": true}')
        run_coroutine(connection_from_.update_state)

        connection_data = run_coroutine(connection_from_.serialize)
        connection_from_.release()
        connection_from_ = None

        connections = AgentConnection.objects.filter(wallet=wallet, id=connection_id, partner_name=partner_name).all()
        if 0 < len(connections):
            connection = connections[0]
            connection.connection_data = json.dumps(connection_data)
            connection.status = 'Active'
        else:
            connection = AgentConnection(
                wallet = wallet,
                partner_name = partner_name,
                invitation = invite_details,
                connection_type = 'Inbound',
                connection_data = json.dumps(connection_data),
                status = 'Active')
        connection.save()

        check_connection_callback(connection, None)
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return connection


def check_connection_status(wallet, connection, initialize_vcx=True):
    """
    Check status of the Connection.
    Called when an invitation has been sent and confirmation has not yet been received.
    Called from the Django background task, but can also be called from a view directly.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and check status
    try:
        prev_status = connection.status
        connection_to_ = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))
        run_coroutine(connection_to_.update_state)
        connection_state = run_coroutine(connection_to_.get_state)
        if connection_state == State.Accepted:
            return_state = 'Active'
        else:
            return_state = 'Sent'

        connection_data = run_coroutine(connection_to_.serialize)
        connection_to_.release()
        connection_to_ = None

        connections = AgentConnection.objects.filter(wallet=wallet, status='Sent', partner_name=connection.partner_name).all()
        my_connection = connections[0]
        my_connection.connection_data = json.dumps(connection_data)
        my_connection.status = return_state
        my_connection.save()

        check_connection_callback(my_connection, prev_status)
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return my_connection


######################################################################
# utilities to offer, request, send and receive credentials
######################################################################

def send_credential_offer(wallet, connection, credential_tag, schema_attrs, cred_def, credential_name, initialize_vcx=True):
    """
    Send a VCX Credential Offer.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        my_connection = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))
        my_cred_def = run_coroutine_with_args(CredentialDef.deserialize, json.loads(cred_def.creddef_data))
        cred_def_handle = my_cred_def.handle

        # create a credential (the last '0' is the 'price')
        credential = run_coroutine_with_args(IssuerCredential.create, credential_tag, schema_attrs, int(cred_def_handle), credential_name, '0')

        run_coroutine_with_args(credential.send_offer, my_connection)

        # serialize/deserialize credential - waiting for Alice to rspond with Credential Request
        credential_data = run_coroutine(credential.serialize)

        conversation = AgentConversation(
            connection = connection,
            conversation_type = 'CredentialOffer',
            message_id = 'N/A',
            status = 'Sent',
            conversation_data = json.dumps(credential_data))
        conversation.save()
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return conversation
    

def send_credential_request(wallet, connection, conversation, initialize_vcx=True):
    """
    Respond to a Credential Offer by sending a VCX Credentia Request.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        my_connection = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))
        #my_offer = run_coroutine_with_args()
    
        offer_json = [json.loads(conversation.conversation_data),]
        credential = run_coroutine_with_args(Credential.create, 'credential', offer_json)

        run_coroutine_with_args(credential.send_request, my_connection, 0)

        # serialize/deserialize credential - wait for Faber to send credential
        credential_data = run_coroutine(credential.serialize)

        conversation.status = 'Sent'
        conversation.conversation_data = json.dumps(credential_data)
        conversation.conversation_type = 'CredentialRequest'
        conversation.save()
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return conversation


######################################################################
# utilities to request, send and receive proofs
######################################################################

def send_proof_request(wallet, connection, proof_uuid, proof_name, proof_attrs, proof_predicates, initialize_vcx=True):
    """
    Send a VCX Proof Request.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        my_connection = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))

        # create a proof request
        proof = run_coroutine_with_kwargs(Proof.create, proof_uuid, proof_name, proof_attrs, {}, requested_predicates=proof_predicates)

        proof_data = run_coroutine(proof.serialize)

        run_coroutine_with_args(proof.request_proof, my_connection)

        # serialize/deserialize credential - waiting for Alice to rspond with Credential Request
        proof_data = run_coroutine(proof.serialize)

        conversation = AgentConversation(
            connection = connection,
            conversation_type = 'ProofRequest',
            message_id = 'N/A',
            status = 'Sent',
            conversation_data = json.dumps(proof_data))
        conversation.save()
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return conversation


# note additional filters are exact match only (attr=value) to filter the allowable claims
# required temporarily because vcx doesn't support the additional filters allowed by indy-sdk
def get_claims_for_proof_request(wallet, connection, my_conversation, additional_filters=None, initialize_vcx=True):
    """
    For the receiver of the Proof Request (i.e. Prover) find the set of claims that can be used
    to construct a Proof.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        my_connection = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))

        # create a proof request
        proof = run_coroutine_with_args(DisclosedProof.create, 'proof', json.loads(my_conversation.conversation_data))

        creds_for_proof = run_coroutine(proof.get_creds)
        # TODO check for filters; remove once this is available in VCX
        if additional_filters and 0 < len(list(additional_filters.keys())):
            ret_creds_for_proof = creds_for_proof.copy()
            for attr in creds_for_proof['attrs']:
                # build array of credential id's (from wallet)
                claims = creds_for_proof['attrs'][attr]
                if 0 < len(claims):
                    ret_claims = []
                    for claim in claims:
                        # TODO this should be filtered using "extra query parameters" in VCX
                        ret = True
                        for key in additional_filters.keys():
                            if claim['cred_info']['attrs'][key] != additional_filters[key]:
                                ret = False
                        if ret:
                            ret_claims.append(claim)
                    ret_creds_for_proof['attrs'][attr] = ret_claims
                else:
                    # if no claim available, we'll request a self-attested value
                    pass
            creds_for_proof = ret_creds_for_proof
    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return creds_for_proof


def cred_for_referent(creds_for_proof, attr, schema_id):
    """
    Find the credential for te given referent (i.e. credential id).
    """

    for cred in creds_for_proof['attrs'][attr]:
        if schema_id == cred['cred_info']['referent']:
            return cred
    return None

def send_claims_for_proof_request(wallet, connection, my_conversation, credential_attrs, initialize_vcx=True):
    """
    Construct a Proof with the given set of claims and send the Proof.
    """

    if initialize_vcx:
        try:
            config_json = wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    # create connection and generate invitation
    try:
        my_connection = run_coroutine_with_args(Connection.deserialize, json.loads(connection.connection_data))

        # load proof request
        proof = run_coroutine_with_args(DisclosedProof.create, 'proof', json.loads(my_conversation.conversation_data))
        creds_for_proof = run_coroutine(proof.get_creds)

        self_attested = {}
        for attr in creds_for_proof['attrs']:
            selected = credential_attrs[attr]
            if 'referent' in selected:
                creds_for_proof['attrs'][attr] = {
                    'credential': cred_for_referent(creds_for_proof, attr, selected['referent'])
                }
            else:
                self_attested[attr] = selected['value']

        for attr in self_attested:
            del creds_for_proof['attrs'][attr]

        # generate and send proof
        run_coroutine_with_args(proof.generate_proof, creds_for_proof, self_attested)
        run_coroutine_with_args(proof.send_proof, my_connection)

        # serialize/deserialize proof 
        proof_data = run_coroutine(proof.serialize)

        my_conversation.status = 'Accepted'
        my_conversation.conversation_type = 'ProofOffer'
        my_conversation.conversation_data = json.dumps(proof_data)
        my_conversation.save()

    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return my_conversation


######################################################################
# utilities to poll for and process outstanding messages
######################################################################

def handle_inbound_messages(my_wallet, my_connection):
    """
    Background task to check for inbound messages.
    Can also be called directly from a view.
    """

    try:
        config_json = my_wallet.wallet_config
        run_coroutine_with_args(vcx_init_with_config, config_json)
    except:
        raise

    try:
        handled_count = 0
        connection_data = json.loads(my_connection.connection_data)
        connection_to_ = run_coroutine_with_args(Connection.deserialize, connection_data)

        if my_connection.connection_type == 'Inbound':
            offers = run_coroutine_with_args(Credential.get_offers, connection_to_)
            for offer in offers:
                already_handled = AgentConversation.objects.filter(message_id=offer[0]['msg_ref_id']).all()
                if len(already_handled) == 0:
                    save_offer = offer[0].copy()
                    offer_data = json.dumps(save_offer)
                    new_offer = AgentConversation(
                                        connection = my_connection,
                                        conversation_type = "CredentialOffer",
                                        message_id = save_offer['msg_ref_id'],
                                        status = 'Pending',
                                        conversation_data = offer_data
                                    )
                    new_offer.save()
                    handled_count = handled_count + 1
                    check_conversation_callback(new_offer, None, None)

        requests = run_coroutine_with_args(DisclosedProof.get_requests, connection_to_)
        for request in requests:
            already_handled = AgentConversation.objects.filter(message_id=request['msg_ref_id']).all()
            if len(already_handled) == 0:
                save_request = request.copy()
                request_data = json.dumps(save_request)
                new_request = AgentConversation(
                                    connection = my_connection,
                                    conversation_type = "ProofRequest",
                                    message_id = save_request['msg_ref_id'],
                                    status = 'Pending',
                                    conversation_data = request_data
                                )
                new_request.save()
                handled_count = handled_count + 1
                check_conversation_callback(new_request, None, None)
    except:
        print("Error polling offers and proof requests")
        # TODO ignore polling errors for now ...
        raise
    finally:
        try:
            shutdown(False)
        except:
            raise

    return handled_count


def poll_message_conversation(my_wallet, my_connection, message, initialize_vcx=True):
    """
    Background task to poll for updates to in-progress Conversations.
    Can also be called directly from a view.
    """

    if initialize_vcx:
        try:
            config_json = my_wallet.wallet_config
            run_coroutine_with_args(vcx_init_with_config, config_json)
        except:
            raise

    try:
        print(" ... Checking message", message.message_id, message.conversation_type)

        connection = run_coroutine_with_args(Connection.deserialize, json.loads(my_connection.connection_data))

        # handle based on message type and status:
        prev_status = message.status
        prev_type = message.conversation_type
        if message.conversation_type == 'CredentialOffer':
            # offer sent from issuer to individual
            # de-serialize message content
            credential = run_coroutine_with_args(IssuerCredential.deserialize, json.loads(message.conversation_data))

            run_coroutine(credential.update_state)
            credential_state = run_coroutine(credential.get_state)

            if credential_state == State.RequestReceived:
                run_coroutine_with_args(credential.send_credential, connection)
                message.conversation_type = 'IssueCredential'
            elif credential_state == State.Accepted:
                message.status = 'Accepted'

            credential_data = run_coroutine(credential.serialize)
            message.conversation_data = json.dumps(credential_data)
            message.save()
        
        elif message.conversation_type == 'CredentialRequest':
            # cred request sent from individual to offerer
            conversation_data_json = json.loads(message.conversation_data)
            credential = run_coroutine_with_args(Credential.deserialize, conversation_data_json)

            run_coroutine(credential.update_state)
            credential_state = run_coroutine(credential.get_state)

            if credential_state == State.Accepted:
                message.status = 'Accepted'

            credential_data = run_coroutine(credential.serialize)
            message.conversation_data = json.dumps(credential_data)
            message.save()

        elif message.conversation_type == 'IssueCredential':
            # credential sent, waiting for acceptance
            # de-serialize message content
            credential = run_coroutine_with_args(IssuerCredential.deserialize, json.loads(message.conversation_data))

            run_coroutine(credential.update_state)
            credential_state = run_coroutine(credential.get_state)

            if credential_state == State.Accepted:
                message.status = 'Accepted'

            # serialize/deserialize credential - wait for Faber to send credential
            credential_data = run_coroutine(credential.serialize)
            message.conversation_data = json.dumps(credential_data)
            message.save()
        
        elif message.conversation_type == 'ProofRequest':
            # proof request send, waiting for proof offer
            # de-serialize message content
            proof = run_coroutine_with_args(Proof.deserialize, json.loads(message.conversation_data))

            run_coroutine(proof.update_state)
            proof_state = run_coroutine(proof.get_state)

            if proof_state == State.Accepted:
                message.status = 'Accepted'
                run_coroutine_with_args(proof.get_proof, connection)

                if proof.proof_state == ProofState.Verified:
                    message.proof_state = 'Verified'
                else:
                    message.proof_state = 'Not Verified'

            # serialize/deserialize credential - wait for Faber to send credential
            print("Saving message with a status of ", message.message_id, message.conversation_type, message.status)
            proof_data = run_coroutine(proof.serialize)
            message.conversation_data = json.dumps(proof_data)
            message.save()

        else:
            print("Error unknown conversation type", message.message_id, message.conversation_type)

        check_conversation_callback(message, prev_type, prev_status)

    except:
        raise
    finally:
        if initialize_vcx:
            try:
                shutdown(False)
            except:
                raise

    return message


def poll_message_conversations(my_wallet, my_connection):
    """
    Background task to poll all Conversations for updates.
    Can also be called directly from a view.
    """

    try:
        config_json = my_wallet.wallet_config
        run_coroutine_with_args(vcx_init_with_config, config_json)
    except:
        raise

    try:
        polled_count = 0

        # Any conversations of status 'Sent' are for bot processing ...
        messages = AgentConversation.objects.filter(connection__wallet=my_wallet, connection=my_connection, status='Sent')

        for message in messages:
            message = poll_message_conversation(my_wallet, my_connection, message, initialize_vcx=False)
            polled_count = polled_count + 1
            pass
    except:
        raise
    finally:
        try:
            shutdown(False)
        except:
            raise

    return polled_count


######################################################################
# optional plug-in call-back for new and updated conversations
######################################################################

def conversation_callback(conversation, prev_type, prev_status):
    """
    Placeholder callback function for Conversation events.
    Override in your application.
    """
    print("conversation callback", prev_type, prev_status, conversation.conversation_type, conversation.status)

def connection_callback(connection, prev_status):
    """
    Placeholder callback function for Connection events.
    Override in your application.
    """
    print("connection callback", prev_status, connection.status)


import importlib

def check_conversation_callback(message, prev_type, prev_status):
    """
    Checks for an application-defined callback for Conversation events and calls it.
    """

    callback_function = getattr(settings, 'INDY_CONVERSATION_CALLBACK', None)
    if callback_function:
        mod_name, func_name = callback_function.rsplit('.',1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)

        func(message, prev_type, prev_status)

def check_connection_callback(connection, prev_status):
    """
    Checks for an application-defined callback for Connection events and calls it.
    """

    callback_function = getattr(settings, 'INDY_CONNECTION_CALLBACK', None)
    if callback_function:
        mod_name, func_name = callback_function.rsplit('.',1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)

        func(connection, prev_status)




# Generated by Django 2.1.7 on 2019-04-13 00:01

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0005_auto_20190409_0051'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='agentconversation',
            name='connection_partner_name',
        ),
        migrations.RemoveField(
            model_name='agentconversation',
            name='wallet',
        ),
        migrations.AddField(
            model_name='agentconversation',
            name='connection',
            field=models.ForeignKey(default='1', on_delete=django.db.models.deletion.CASCADE, to='indy_community.AgentConnection', to_field='partner_name'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='agentconnection',
            name='partner_name',
            field=models.CharField(max_length=60, unique=True),
        ),
    ]



import asyncio
from pathlib import Path
from tempfile import gettempdir
from os import environ
import asyncio
import aiohttp

from django.apps import AppConfig
from ctypes import *

import time

import json

import sys
from ctypes import *

from indy import anoncreds, crypto, did, ledger, pool, wallet
from indy.error import ErrorCode, IndyError

from django.conf import settings

from .indy_state import set_pool_handle


PROTOCOL_VERSION = 2


def path_home() -> Path:
    return Path.home().joinpath(".indy_client")


async def _fetch_url(the_url):
    async with aiohttp.ClientSession() as session:
        async with session.get(the_url) as resp:
            r_status = resp.status
            r_text = await resp.text()
            return (r_status, r_text)

async def _fetch_genesis_txn(genesis_url: str) -> bool:
    try:
        (r_status, data) = await _fetch_url(genesis_url)
    except:
        raise

    # check data is valid json
    lines = data.splitlines()
    if not lines or not json.loads(lines[0]):
        raise Exception("Genesis transaction file is not valid JSON")

    return data


# this is the genesis that connects to the default indy-sdk ledger, checkout indy-sdk and run:
#    docker build -f ci/indy-pool.dockerfile -t indy_pool .
#    docker run -itd -p 9701-9708:9701-9708 indy_pool
async def pool_genesis_txn_data():
    if 'vcx_genesis_url' in settings.INDY_CONFIG:
        # download from the genesis url, if specified
        genesis_url = settings.INDY_CONFIG['vcx_genesis_url']
        genesis_txn = await _fetch_genesis_txn(genesis_url)
    else:
        # this is the default if not specified
        pool_ip = environ.get("TEST_POOL_IP", "127.0.0.1")

        genesis_txn = "\n".join([
            '{{"reqSignature":{{}},"txn":{{"data":{{"data":{{"alias":"Node1","blskey":"4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba","blskey_pop":"RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1","client_ip":"{}","client_port":9702,"node_ip":"{}","node_port":9701,"services":["VALIDATOR"]}},"dest":"Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv"}},"metadata":{{"from":"Th7MpTaRZVRYnPiabds81Y"}},"type":"0"}},"txnMetadata":{{"seqNo":1,"txnId":"fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62"}},"ver":"1"}}'.format(
                pool_ip, pool_ip),
            '{{"reqSignature":{{}},"txn":{{"data":{{"data":{{"alias":"Node2","blskey":"37rAPpXVoxzKhz7d9gkUe52XuXryuLXoM6P6LbWDB7LSbG62Lsb33sfG7zqS8TK1MXwuCHj1FKNzVpsnafmqLG1vXN88rt38mNFs9TENzm4QHdBzsvCuoBnPH7rpYYDo9DZNJePaDvRvqJKByCabubJz3XXKbEeshzpz4Ma5QYpJqjk","blskey_pop":"Qr658mWZ2YC8JXGXwMDQTzuZCWF7NK9EwxphGmcBvCh6ybUuLxbG65nsX4JvD4SPNtkJ2w9ug1yLTj6fgmuDg41TgECXjLCij3RMsV8CwewBVgVN67wsA45DFWvqvLtu4rjNnE9JbdFTc1Z4WCPA3Xan44K1HoHAq9EVeaRYs8zoF5","client_ip":"{}","client_port":9704,"node_ip":"{}","node_port":9703,"services":["VALIDATOR"]}},"dest":"8ECVSk179mjsjKRLWiQtssMLgp6EPhWXtaYyStWPSGAb"}},"metadata":{{"from":"EbP4aYNeTHL6q385GuVpRV"}},"type":"0"}},"txnMetadata":{{"seqNo":2,"txnId":"1ac8aece2a18ced660fef8694b61aac3af08ba875ce3026a160acbc3a3af35fc"}},"ver":"1"}}'.format(
                pool_ip, pool_ip),
            '{{"reqSignature":{{}},"txn":{{"data":{{"data":{{"alias":"Node3","blskey":"3WFpdbg7C5cnLYZwFZevJqhubkFALBfCBBok15GdrKMUhUjGsk3jV6QKj6MZgEubF7oqCafxNdkm7eswgA4sdKTRc82tLGzZBd6vNqU8dupzup6uYUf32KTHTPQbuUM8Yk4QFXjEf2Usu2TJcNkdgpyeUSX42u5LqdDDpNSWUK5deC5","blskey_pop":"QwDeb2CkNSx6r8QC8vGQK3GRv7Yndn84TGNijX8YXHPiagXajyfTjoR87rXUu4G4QLk2cF8NNyqWiYMus1623dELWwx57rLCFqGh7N4ZRbGDRP4fnVcaKg1BcUxQ866Ven4gw8y4N56S5HzxXNBZtLYmhGHvDtk6PFkFwCvxYrNYjh","client_ip":"{}","client_port":9706,"node_ip":"{}","node_port":9705,"services":["VALIDATOR"]}},"dest":"DKVxG2fXXTU8yT5N7hGEbXB3dfdAnYv1JczDUHpmDxya"}},"metadata":{{"from":"4cU41vWW82ArfxJxHkzXPG"}},"type":"0"}},"txnMetadata":{{"seqNo":3,"txnId":"7e9f355dffa78ed24668f0e0e369fd8c224076571c51e2ea8be5f26479edebe4"}},"ver":"1"}}'.format(
                pool_ip, pool_ip),
            '{{"reqSignature":{{}},"txn":{{"data":{{"data":{{"alias":"Node4","blskey":"2zN3bHM1m4rLz54MJHYSwvqzPchYp8jkHswveCLAEJVcX6Mm1wHQD1SkPYMzUDTZvWvhuE6VNAkK3KxVeEmsanSmvjVkReDeBEMxeDaayjcZjFGPydyey1qxBHmTvAnBKoPydvuTAqx5f7YNNRAdeLmUi99gERUU7TD8KfAa6MpQ9bw","blskey_pop":"RPLagxaR5xdimFzwmzYnz4ZhWtYQEj8iR5ZU53T2gitPCyCHQneUn2Huc4oeLd2B2HzkGnjAff4hWTJT6C7qHYB1Mv2wU5iHHGFWkhnTX9WsEAbunJCV2qcaXScKj4tTfvdDKfLiVuU2av6hbsMztirRze7LvYBkRHV3tGwyCptsrP","client_ip":"{}","client_port":9708,"node_ip":"{}","node_port":9707,"services":["VALIDATOR"]}},"dest":"4PS3EDQ3dW1tci1Bp6543CfuuebjFrg36kLAUcskGfaA"}},"metadata":{{"from":"TWwCRQRZ2ZHMJFn9TzLp7W"}},"type":"0"}},"txnMetadata":{{"seqNo":4,"txnId":"aa5e817d7cc626170eca175822029339a444eb0ee8f0bd20d3b0b76e566fb008"}},"ver":"1"}}'.format(
                pool_ip, pool_ip)
        ])

    indy_config = getattr(settings, 'INDY_CONFIG')
    f = open(indy_config['vcx_genesis_path'], "w+")
    f.write(genesis_txn)
    f.close()

    return genesis_txn

def run_coroutine(coroutine):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coroutine())
    finally:
        loop.close()

async def get_pool_genesis_txn_path(pool_name):
    path_temp = Path(gettempdir()).joinpath("indy")
    path = path_temp.joinpath("{}.txn".format(pool_name))
    await save_pool_genesis_txn_file(path)
    return path

async def save_pool_genesis_txn_file(path):
    data = await pool_genesis_txn_data()

    path.parent.mkdir(parents=True, exist_ok=True)

    with open(str(path), "w+") as f:
        f.writelines(data)


class IndyCoreConfig(AppConfig):
    name = 'indy_community'

    def ready(self):
        # import login/logout signals
        import indy_community.signals

        pg_dll = settings.INDY_CONFIG['storage_dll']
        pg_entrypoint = settings.INDY_CONFIG['storage_entrypoint']
        print('Loading {}'.format(pg_dll))
        stg_lib = CDLL(pg_dll)
        result = stg_lib[pg_entrypoint]()
        if result != 0:
            print('Error unable to load wallet storage {}'.format(result))
            raise AppError('Error unable to load wallet storage {}'.format(result))

        pay_dll = settings.INDY_CONFIG['payment_dll']
        pay_entrypoint = settings.INDY_CONFIG['payment_entrypoint']
        print('Loading {}'.format(pay_dll))
        pay_lib = CDLL(pay_dll)
        result = pay_lib[pay_entrypoint]()
        if result != 0:
            print('Error unable to load payment plug-in {}'.format(result))
            raise AppError('Error unable to load payment plug-in {}'.format(result))

        pool_handle = run_coroutine(run)
        set_pool_handle(pool_handle)
        time.sleep(1)  # FIXME waiting for libindy thread complete

        print("App is ready!!!")


async def run():
    print("Getting started -> started")

    pool_ = {
        'name': 'pool1'
    }
    print("Open Pool Ledger: {}".format(pool_['name']))
    pool_['genesis_txn_path'] = await get_pool_genesis_txn_path(pool_['name'])
    print(pool_['genesis_txn_path'])
    pool_['config'] = json.dumps({"genesis_txn": str(pool_['genesis_txn_path'])})

    # Set protocol version 2 to work with Indy Node 1.4
    await pool.set_protocol_version(PROTOCOL_VERSION)

    try:
        await pool.create_pool_ledger_config(pool_['name'], pool_['config'])
    except IndyError as ex:
        if ex.error_code == ErrorCode.PoolLedgerConfigAlreadyExistsError:
            pass
    pool_handle = await pool.open_pool_ledger(pool_['name'], None)
    print("Returned pool handle", pool_handle)
    return pool_handle




from django.core.management.base import BaseCommand
from django.utils import timezone
import yaml
import os

from indy_community.models import *
from indy_community.agent_utils import *


class Command(BaseCommand):
    help = 'Loads Schemas and optionally creates Credential Definitions'

    def add_arguments(self, parser):
        parser.add_argument('config_file', nargs='+')
        parser.add_argument('org_id', nargs='+', type=int)

        # Named (optional) arguments
        parser.add_argument(
            '--cred_defs',
            action='store_true',
            dest='cred_defs',
            help='Create Credential Definitions for the selected organization, for all Schemas',
        )

    def handle(self, *args, **options):
        schemas = None
        org = None

        # verify that config file exists and we can load yaml
        self.stdout.write("config_file = %s" % str(options['config_file']))
        with open(str(options['config_file'][0]), 'r') as stream:
            try:
                schemas = yaml.load(stream)
            except yaml.YAMLError as exc:
                self.stdout.write(exc)
                raise

            # validate data in yaml file
            for name in schemas:
                spec = schemas[name]
                # TODO validation

        # verify org exists
        self.stdout.write("issue schemas for org_id = %s" % str(options['org_id'][0]))
        orgs = IndyOrganization.objects.filter(id=options['org_id'][0]).all()
        if 0 == len(orgs):
            self.stdout.write("no organization found for org_id = %s" % str(options['org_id']))
        else:
            org = orgs[0]

        # now create schemas (and potentially create cred defs)
        if schemas and org:
            for name in schemas:
                spec = schemas[name]

                if spec['type'] == 'schema':
                    if spec['version'] == '$generate':
                        version = random_schema_version()
                    else:
                        version = spec['version']
                    attributes = spec['attributes']
                    # TODO save roles (to auto-create cred defs for the schema(s))
                    (schema_json, creddef_template) = create_schema_json(name, version, attributes)
                    schema = create_schema(org.wallet, schema_json, creddef_template)
                    self.stdout.write("Created schema for %s" % name)

                    # add role(s) to this schema
                    if 'issuing_roles' in spec:
                        for role_name in spec['issuing_roles']:
                            role, created = IndyOrgRole.objects.get_or_create(name=role_name)
                            schema.roles.add(role)
                        schema.save()

                    # add cred def?
                    if options['cred_defs']:
                        creddef = create_creddef(org.wallet, schema, schema.schema_name + '-' + org.wallet.wallet_name, schema.schema_template)
                        self.stdout.write("Created cred def for schema %s" % name)
                elif spec['type'] == 'proof_request':
                    description = spec['description']
                    if 'revealed_attributes' in spec:
                        revealed_attributes = spec['revealed_attributes']
                    else:
                        revealed_attributes = []
                    if 'predicates' in spec:
                        predicates = spec['predicates']
                    else:
                        predicates = []
                    create_proof_request(name, description, revealed_attributes, predicates)
                    self.stdout.write("Created proof request for %s" % name)




from django.conf import settings
from django.db import models
from django.contrib.sessions.models import Session
from django.contrib.auth.base_user import AbstractBaseUser, BaseUserManager
from django.contrib.auth.models import Group, PermissionsMixin

from django.utils import timezone

from datetime import datetime, date, timedelta
import json


USER_ROLES = (
    'Admin',
    'User',
)

# base class for Indy wallets
class IndyWallet(models.Model):
    wallet_name = models.CharField(max_length=60, unique=True)
    wallet_config = models.TextField(max_length=4000, blank=True)

    def __str__(self):
        return self.wallet_name

# special class for managing Indy users and wallets
class IndyUserManager(BaseUserManager):

    def _create_user(self, email, password, **extra_fields):
        if not email:
            raise ValueError("The given email must be set")
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)

        user.set_password(password)
        user.save()

        return user

    def create_user(self, email, password=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)

        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get('is_superuser') is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return self._create_user(email, password, **extra_fields)

# special class for Indy users (owns a wallet)
class IndyUser(AbstractBaseUser, PermissionsMixin):
    """
    Simple custom User class with email-based login.
    """
    email = models.EmailField(unique=True)
    first_name = models.CharField(max_length=80, blank=True)
    last_name = models.CharField(max_length=150, blank=True)
    is_staff = models.BooleanField(
        default=False,
        help_text="Designates whether the user can log into this admin site."
    )
    is_active = models.BooleanField(
        default=True,
        help_text=(
            "Designates whether this user should be treated as active. "
            "Unselect this instead of deleting accounts."
        )
    )
    date_joined = models.DateTimeField(default=timezone.now)
    last_login = models.DateTimeField(blank=True, null=True)
    wallet = models.ForeignKey(IndyWallet, to_field="wallet_name", related_name='wallet_user', blank = True, null=True, on_delete=models.CASCADE)
    managed_wallet = models.BooleanField(default=True)

    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'

    objects = IndyUserManager()

    @property
    def roles(self):
        # -> Iterable
        # Produce a list of the given user's roles.

        return filter(self.has_role, USER_ROLES)

    def get_full_name(self):
        return "%s %s" % (self.first_name, self.last_name)

    def add_role(self, role):
        # String ->
        # Adds user to role group

        self.groups.add(Group.objects.get(name=role))

    def has_role(self, role):
        # String -> Boolean
        # Produce true if user is in the given role group.

        return self.groups.filter(name=role).exists()


# Roles to which an organization can belong
class IndyOrgRole(models.Model):
    name = models.CharField(max_length=40, unique=True)

    def __str__(self):
        return self.name

# Base class for organizations that use the Indy platform
class IndyOrganization(models.Model):
    org_name = models.CharField(max_length=60, unique=True)
    wallet = models.ForeignKey(IndyWallet, to_field="wallet_name", related_name='wallet_org', blank = True, null=True, on_delete=models.CASCADE)
    role = models.ForeignKey(IndyOrgRole, blank = True, null=True, on_delete=models.CASCADE)
    ico_url = models.CharField(max_length=120, blank = True, null=True)
    managed_wallet = models.BooleanField(default=True)

    def __str__(self):
        return self.org_name

# Association class for user/organization relationship
class IndyOrgRelationship(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='indyrelationship_set', on_delete=models.CASCADE)
    org = models.ForeignKey(IndyOrganization, related_name='indyrelationship_set', on_delete=models.CASCADE)

    def __str__(self):
        return self.user.email + ':' + self.org.org_name

# track user session and attached wallet for background agents
class IndySession(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    session = models.ForeignKey(Session, on_delete=models.CASCADE)
    wallet_name = models.CharField(max_length=60, blank=True, null=True)


# reference to a schema on the ledger
class IndySchema(models.Model):
    ledger_schema_id = models.CharField(max_length=80, unique=True)
    schema_name = models.CharField(max_length=80)
    schema_version = models.CharField(max_length=80)
    schema = models.TextField(max_length=4000)
    schema_template = models.TextField(max_length=4000)
    schema_data = models.TextField(max_length=4000)
    # orgs which contain these role(s) will automatically get cred defs created for this schema
    roles = models.ManyToManyField(IndyOrgRole)

    def __str__(self):
        return self.schema_name


# reference to a credential definition on the ledger
class IndyCredentialDefinition(models.Model):
    ledger_creddef_id = models.CharField(max_length=80, unique=True)
    ledger_schema = models.ForeignKey(IndySchema, on_delete=models.CASCADE)
    wallet = models.ForeignKey(IndyWallet, to_field="wallet_name", related_name='indycreddef_set', on_delete=models.CASCADE)
    creddef_name = models.CharField(max_length=80)
    creddef_handle = models.CharField(max_length=80)
    creddef_template = models.TextField(max_length=4000)
    creddef_data = models.TextField(max_length=4000)

    def __str__(self):
        return self.ledger_schema.schema_name + ":" + self.wallet.wallet_name + ":" + self.creddef_name


# Description of a proof request
class IndyProofRequest(models.Model):
    proof_req_name = models.CharField(max_length=400, unique=True)
    proof_req_description = models.TextField(max_length=4000)
    proof_req_attrs = models.TextField(max_length=4000)
    proof_req_predicates = models.TextField(max_length=4000, blank=True)

    def __str__(self):
        return self.proof_req_name


# base class for Agent connections
class AgentConnection(models.Model):
    wallet = models.ForeignKey(IndyWallet, to_field="wallet_name", on_delete=models.CASCADE)
    partner_name = models.CharField(max_length=60)
    invitation = models.TextField(max_length=4000, blank=True)
    token = models.CharField(max_length=80, blank=True)
    status = models.CharField(max_length=20)
    connection_type = models.CharField(max_length=20)
    connection_data = models.TextField(max_length=4000, blank=True)

    def __str__(self):
        return self.wallet.wallet_name + ":" + self.partner_name + ", " +  self.status

    # script from @burdettadam, map to the invite format expected by Connect.Me
    def invitation_shortform(self, source_name, target_name, institution_logo_url):
        invite = json.loads(self.invitation)
        cm_invite = { "id": invite["connReqId"],
                "s" :{"d" :invite["senderDetail"]["DID"],
                        "dp":{"d":invite["senderDetail"]["agentKeyDlgProof"]["agentDID"],
                              "k":invite["senderDetail"]["agentKeyDlgProof"]["agentDelegatedKey"],
                              "s":invite["senderDetail"]["agentKeyDlgProof"]["signature"]
                            },
                        "l" :invite["senderDetail"]["logoUrl"],
                        "n" :invite["senderDetail"]["name"],
                        "v" :invite["senderDetail"]["verKey"]
                        },
                "sa":{"d":invite["senderAgencyDetail"]["DID"],
                        "e":invite["senderAgencyDetail"]["endpoint"],
                        "v":invite["senderAgencyDetail"]["verKey"]
                    },
                "sc":invite["statusCode"],
                "sm":invite["statusMsg"],
                "t" :invite["targetName"]
                }
        cm_invite["s"]["n"] = source_name
        cm_invite["t"] = target_name
        cm_invite["s"]["l"] = institution_logo_url
        return json.dumps(cm_invite)


# base class for Agent conversations - issue/receive credential and request/provide proof
class AgentConversation(models.Model):
    connection = models.ForeignKey(AgentConnection, on_delete=models.CASCADE)
    conversation_type = models.CharField(max_length=30)
    message_id = models.CharField(max_length=30)
    status = models.CharField(max_length=20)
    proof_state = models.CharField(max_length=20, blank=True)
    conversation_data = models.TextField(max_length=4000, blank=True)

    def __str__(self):
        return self.connection.wallet.wallet_name + ":" + self.connection.partner_name + ":" + self.message_id + ", " +  self.conversation_type + " " + self.status



from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.views.generic.edit import UpdateView

from ..models import *
from ..wallet_utils import *
from ..registration_utils import *


class RegistrationTests(TestCase):

    def test_user_registration(self):
        # create, register and provision a user
        email = 'test1@registration.com'
        user = get_user_model().objects.create(
            email=email,
            first_name='Test',
            last_name='Registration',
        )
        user.save()
        raw_password = 'pass1234'
        user_provision(user, raw_password)

        # fetch and open wallet
        fetch_user = get_user_model().objects.filter(email=email).all()[0]
        user_wallet_name = get_user_wallet_name(email)
        self.assertEqual(fetch_user.wallet.wallet_name, user_wallet_name)
        wallet_handle = open_wallet(user_wallet_name, raw_password)
        res = close_wallet(wallet_handle)
        self.assertEqual(res, 0)

        # cleanup after ourselves
        res = delete_wallet(user_wallet_name, raw_password)
        self.assertEqual(res, 0)


    def test_org_registration(self):
        # create, register and provision a user and org
        # create, register and provision a user
        email = 'test2@registration.com'
        user_wallet_name = get_user_wallet_name(email)
        user = get_user_model().objects.create(
            email=email,
            first_name='Test',
            last_name='Registration',
        )
        user.save()
        raw_password = 'pass1234'
        user_provision(user, raw_password)

        # now org
        org_name = 'My Unittest Org Inc'
        org = org_signup(user, raw_password, org_name)

        # fetch everything ...
        fetch_user = get_user_model().objects.filter(email=email).all()[0]
        fetch_org = IndyOrganization.objects.filter(org_name=org_name).all()[0]
        org_wallet_name = get_org_wallet_name(org_name)
        self.assertEqual(fetch_org.wallet.wallet_name, org_wallet_name)
        wallet_handle = open_wallet(org_wallet_name, raw_password)
        res = close_wallet(wallet_handle)
        self.assertEqual(res, 0)

        # verify relationship
        self.assertEqual(len(fetch_user.indyrelationship_set.all()), 1)
        user_org = fetch_user.indyrelationship_set.all()[0].org
        self.assertEqual(user_org.org_name, org_name)

        self.assertEqual(len(fetch_org.indyrelationship_set.all()), 1)
        org_user = fetch_org.indyrelationship_set.all()[0].user
        self.assertEqual(org_user.email, email)

        # cleanup after ourselves
        res = delete_wallet(org_wallet_name, raw_password)
        self.assertEqual(res, 0)
        res = delete_wallet(user_wallet_name, raw_password)
        self.assertEqual(res, 0)




import json
import datetime
import pytz

from django.utils import timezone
from django.contrib.sessions.models import Session
from django.contrib.auth import get_user_model

from background_task import background

from indy.error import ErrorCode, IndyError

from .models import IndySession, IndyWallet, AgentConnection, AgentConversation
from .agent_utils import check_connection_status, handle_inbound_messages, poll_message_conversations

AGENT_POLL_INTERVAL = 5


@background(schedule=AGENT_POLL_INTERVAL)
def agent_background_task(message, user_id, session_key, org_id=None):
    # check if user/wallet has a valid session
    user = get_user_model().objects.filter(id=user_id).first()
    try:
        session = IndySession.objects.get(user=user, session_id=session_key)
    except:
        raise Exception("No Indy Session found for {} {}".format(user.email, session_key))

    print("Found session {}  for user {} wallet {}".format(session.id, user.email, session.wallet_name))
    if session.session.expire_date < timezone.get_current_timezone().localize(datetime.datetime.now()):
        raise Exception("Django Session timed out for {} {}".format(user.email, session.wallet_name))

    if session.wallet_name is not None:
        wallet = IndyWallet.objects.get(wallet_name=session.wallet_name)

        # check for outstanding connections and poll status
        connections = AgentConnection.objects.filter(wallet=wallet, status='Sent').all()

        # if (anything to do) initialize VCX agent and do all our updates
        # TODO (for now each request re-initializes VCX)
        for connection in connections:
            # validate connection and get the updated status
            try:
                upd_connection = check_connection_status(wallet, connection)
            except IndyError as e:
                print(" >>> Failed to update connection request for", session.wallet_name, connection.id, connection.partner_name)
                raise e

        # check for outstanding connections and poll status
        connections = AgentConnection.objects.filter(wallet=wallet, status='Active').all()
        for connection in connections:
            # check for outstanding, un-received messages - add to outstanding conversations
            try:
                #if connection.connection_type == 'Inbound':
                msg_count = handle_inbound_messages(wallet, connection)
            except IndyError as e:
                print(" >>> Failed to handle inbound messages for", session.wallet_name, connection.id, connection.partner_name)
                raise e

            # check status of any in-flight conversations (send/receive credential or request/provide proof)
            try:
                polled_count = poll_message_conversations(wallet, connection)
            except IndyError as e:
                print(" >>> Failed to poll conversations for", session.wallet_name, connection.id, connection.partner_name)
                raise e





from .models import *
from .utils import *
from .indy_utils import *
from .wallet_utils import *
from .agent_utils import *


def user_provision(user, raw_password):
    """
    Create a new user wallet and associate with the user
    """

    wallet_name = get_user_wallet_name(user.email)
    res = create_wallet(wallet_name, raw_password)
    if res != 0:
        raise Exception("Error wallet create failed: " + str(res))

    # provision as an agent wallet (errors will raise exceptions)
    config = initialize_and_provision_vcx(wallet_name, raw_password, user.email)

    # save everything to our database
    wallet = IndyWallet.objects.create(wallet_name=wallet_name, wallet_config = config)
    wallet.save()
    user.wallet = wallet
    user.save()

    return user


def org_provision(org, raw_password, org_role=None):
    """
    Create a new org wallet and associate to the org
    """

    wallet_name = get_org_wallet_name(org.org_name)
    res = create_wallet(wallet_name, raw_password)
    if res != 0:
        raise Exception("Error wallet create failed: " + str(res))

    # create a did for this org
    did_seed = calc_wallet_seed(wallet_name)
    if org_role != "Trustee":
        create_and_register_did(wallet_name, org_role)

    # provision as an agent wallet (errors will raise exceptions)
    config = initialize_and_provision_vcx(wallet_name, raw_password, org.org_name, did_seed=did_seed, org_role=org_role, institution_logo_url=org.ico_url)

    # save everything to our database
    wallet = IndyWallet.objects.create(wallet_name=wallet_name, wallet_config = config)
    wallet.save()
    org.wallet = wallet
    org.save()

    # if the org has a role, check if there are any schemas associated with that role
    if org_role:
        role_schemas = IndySchema.objects.filter(roles=org_role).all()
        for schema in role_schemas:
            creddef = create_creddef(org.wallet, schema, schema.schema_name + '-' + org.wallet.wallet_name, schema.schema_template)

    return org


def org_signup(user, raw_password, org_name, org_attrs={}, org_relation_attrs={}, org_role=None, org_ico_url=None):
    """
    Helper method to create and provision a new org, and associate to the current user
    """
    
    if not org_ico_url:
        org_ico_url = 'http://robohash.org/456'

    org = get_indy_settings_model('INDY_ORGANIZATION_MODEL').objects.create(org_name=org_name, role=org_role, ico_url=org_ico_url, **org_attrs)

    org = org_provision(org, raw_password, org_role)

    # associate the user with the org
    relation = get_indy_settings_model('INDY_ORG_RELATION_MODEL').objects.create(org=org, user=user, **org_relation_attrs)
    relation.save()

    return org




# Generated by Django 2.1.7 on 2019-03-19 22:07

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='indycredentialdefinition',
            name='wallet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='indycreddef_set', to='indy_community.IndyWallet', to_field='wallet_name'),
        ),
        migrations.AlterField(
            model_name='indyorgrelationship',
            name='org',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='indyrelationship_set', to='indy_community.IndyOrganization'),
        ),
        migrations.AlterField(
            model_name='indyorgrelationship',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='indyrelationship_set', to='indy_community.IndyUser'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-04-30 01:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('indy_community', '0007_auto_20190413_0125'),
    ]

    operations = [
        migrations.AddField(
            model_name='indyorganization',
            name='ico_url',
            field=models.CharField(blank=True, max_length=120, null=True),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, Client, modify_settings, override_settings
from django.core import management
from django.core.management.base import CommandError
from django.utils.six import StringIO

from ..utils import *
from ..wallet_utils import *
from ..models import *
from ..views import *
from ..forms import *
import indy_community.management.commands.loads_schemas

User = get_user_model()


class IndyViewsTests(TestCase):
    PASSWORD = 'exampl12345'

    def login_user(self, user_name, raw_password):
        url = reverse('login')
        resp = self.client.post(url, {'username': user_name, 'password': raw_password, 'next': '/individual/profile/'}, follow=True)
        self.assertEqual(resp.status_code, 200)

    def login_org_user(self, user_name, raw_password):
        url = reverse('login')
        resp = self.client.post(url, {'username': user_name, 'password': raw_password, 'next': '/organization/profile/'}, follow=True)
        self.assertEqual(resp.status_code, 200)

    def register_and_login_user(self, user_name, raw_password, first_name, last_name):
        # register
        url = reverse('signup')
        resp = self.client.get(url, follow=True)
        self.assertEqual(resp.status_code, 200)

        data = {'first_name': first_name, 'last_name': last_name, 'email': user_name, 
                'password1': raw_password, 'password2': raw_password}
        form = UserSignUpForm(data=data)
        self.assertTrue(form.is_valid())

        url = reverse('signup')
        resp = self.client.post(url, data, follow=True)

        self.login_user(user_name, raw_password)

    def register_and_login_org(self, user_name, raw_password, first_name, last_name, org_name, org_role):
        # register
        url = reverse('signup')
        resp = self.client.get(url, follow=True)
        self.assertEqual(resp.status_code, 200)

        data = {'first_name': first_name, 'last_name': last_name, 'email': user_name, 
                'password1': raw_password, 'password2': raw_password,
                'org_name': org_name, 'org_role_name': org_role}
        form = OrganizationSignUpForm(data=data)
        self.assertTrue(form.is_valid())

        url = reverse('org_signup')
        resp = self.client.post(url, data, follow=True)

        self.login_org_user(user_name, raw_password)

    def logout_user(self):
        url = reverse('logout')
        resp = self.client.post(url, follow=True)

    def cleanup_user(self, user, raw_password):
        self.logout_user()

        # cleanup after ourselves
        user_wallet_name = user.wallet.wallet_name
        res = delete_wallet(user_wallet_name, raw_password)
        self.assertEqual(res, 0)

    def cleanup_user_and_org(self, user, org, raw_password):
        self.logout_user()

        # cleanup after ourselves
        org_wallet_name = org.wallet.wallet_name
        res = delete_wallet(org_wallet_name, raw_password)
        self.assertEqual(res, 0)


    def test_register_user_and_login(self):
        USER_USER = 'user_' + random_alpha_string(10) + '@mail.com'

        self.register_and_login_user(USER_USER, self.PASSWORD, 'Random', 'Name')

        # asserts
        url = reverse('individual:profile')
        resp = self.client.get(url, follow=True)
        self.assertEqual(resp.status_code, 200)

        fetch_users = User.objects.filter(email=USER_USER).all()
        self.assertEqual(1, len(fetch_users))

        self.cleanup_user(fetch_users[0], self.PASSWORD)


    def test_register_org_user_and_login(self):
        ORG_USER = 'org_' + random_alpha_string(10) + '@mail.com'
        ORG_NAME = 'Org ' + random_alpha_string(10) + ' Inc'

        self.register_and_login_org(ORG_USER, self.PASSWORD, 'Random', 'Name', ORG_NAME, 'Trustee')
        
        # asserts
        url = reverse('organization:profile')
        resp = self.client.get(url, follow=True)
        self.assertEqual(resp.status_code, 200)

        fetch_users = User.objects.filter(email=ORG_USER).all()
        self.assertEqual(1, len(fetch_users))
        fetch_orgs = IndyOrganization.objects.filter(org_name=ORG_NAME).all()
        self.assertEqual(1, len(fetch_orgs))

        self.cleanup_user_and_org(fetch_users[0], fetch_orgs[0], self.PASSWORD)

    def test_register_orgs_with_schemas(self):
        ORG_USER = 'org_' + random_alpha_string(10) + '@mail.com'
        ORG_NAME = 'Org ' + random_alpha_string(10) + ' Inc'

        self.register_and_login_org(ORG_USER, self.PASSWORD, 'Random', 'Name', ORG_NAME, 'Trustee')
        self.logout_user()

        fetch_users = User.objects.filter(email=ORG_USER).all()
        self.assertEqual(1, len(fetch_users))
        fetch_orgs = IndyOrganization.objects.filter(org_name=ORG_NAME).all()
        self.assertEqual(1, len(fetch_orgs))

        # load schemas and cred defs from yml config
        config = './imms-schemas.yml'
        out = StringIO()
        management.call_command('loads_schemas', config, str(fetch_orgs[0].id), stdout=out)

        schemas = IndySchema.objects.all()
        self.assertEqual(4, len(schemas))
        proofs = IndyProofRequest.objects.all()
        self.assertEqual(4, len(proofs))

        # load orgs and cred defs from yml config
        config = './imms-orgs.yml'
        out = StringIO()
        management.call_command('loads_orgs', config, stdout=out)

        orgs = IndyOrganization.objects.all()
        self.assertEqual(4, len(orgs))
        cred_defs = IndyCredentialDefinition.objects.all()
        self.assertEqual(4, len(cred_defs))

        # TODO fetch the org for each role

        # TODO register a user Alice

        # TODO issue credentials and proofs ...

        # TODO cleanup - delete all wallets

        self.login_org_user(ORG_USER, self.PASSWORD)
        self.cleanup_user_and_org(fetch_users[0], fetch_orgs[0], self.PASSWORD)




from django.http import HttpResponseBadRequest, HttpResponseRedirect, HttpResponse
from django.shortcuts import render, redirect
from django.contrib.auth import authenticate, get_user_model, login
from django.urls import reverse
from django.conf import settings

import pyqrcode
#import qrcode
import uuid

from .forms import *
from .models import *
from .wallet_utils import *
from .registration_utils import *
from .agent_utils import *
from .signals import handle_wallet_login_internal


USER_ROLE = getattr(settings, "DEFAULT_USER_ROLE", 'User')
ORG_ROLE = getattr(settings, "DEFAULT_ORG_ROLE", 'Admin')

###############################################################
# UI views to support user and organization registration
###############################################################
def mobile_request_connection(
    request,
    form_template='registration/request_mobile_connection.html',
    response_template='registration/mobile_connection_info.html'
    ):
    """
    Create an invitation for a user who will use a mobile wallet.
    No managed wallet is created.  A login account is created but many of the built-in screens
    won't work due to the missing wallet.
    """

    # user requests mobile connection to an org
    if request.method == 'POST':
        # generate ivitation and display a QR code
        form = RequestMobileConnectionForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            # first save a local user with a non-managed wallet
            cd = form.cleaned_data
            form.save()
            username = cd.get('email')
            raw_password = cd.get('password1')
            user = authenticate(username=username, password=raw_password)
            user.managed_wallet = False

            if Group.objects.filter(name=USER_ROLE).exists():
                user.groups.add(Group.objects.get(name=USER_ROLE))
            user.save()

            org = cd.get('org')
            email = cd.get('email')
            partner_name = email + ' (mobile)'

            # get requested org and their wallet
            org_wallet = org.wallet

            # mobile user not registered locally
            target_user = None
            their_wallet = None

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            # build the connection and get the invitation data back
            try:
                org_connection = send_connection_invitation(org_wallet, partner_name)

                return render(request, 'registration/mobile_connection_info.html', {'org_name': org.org_name, 'connection_token': org_connection.token})
            except Exception as e:
                # ignore errors for now
                print(" >>> Failed to create request for", org_wallet.wallet_name)
                print(e)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to create request for ' + org_wallet.wallet_name})

    else:
        # populate form and get info from user
        form = RequestMobileConnectionForm(initial={})
        return render(request, 'registration/request_mobile_connection.html', {'form': form})


# Sign up as a site user, and create a wallet
def user_signup_view(
    request,
    template=''
    ):
    """
    Create a user account with a managed wallet.
    """

    if request.method == 'POST':
        form = UserSignUpForm(request.POST)
        if form.is_valid():
            form.save()
            username = form.cleaned_data.get('email')
            raw_password = form.cleaned_data.get('password1')
            user = authenticate(username=username, password=raw_password)

            if Group.objects.filter(name=USER_ROLE).exists():
                user.groups.add(Group.objects.get(name=USER_ROLE))
            user.save()

            # create an Indy wallet - derive wallet name from email, and re-use raw password
            user = user_provision(user, raw_password)

            # TODO need to auto-login with Atria custom user
            #login(request, user)

            return redirect('login')
    else:
        form = UserSignUpForm()
    return render(request, 'registration/signup.html', {'form': form})


# Sign up as an org user, and create a wallet
def org_signup_view(
    request,
    template=''
    ):
    """
    Signup an Organization with a managed wallet.
    Creates a user account and links to the Organization.
    """

    if request.method == 'POST':
        form = OrganizationSignUpForm(request.POST)
        if form.is_valid():
            form.save()
            username = form.cleaned_data.get('email')
            raw_password = form.cleaned_data.get('password1')
            user = authenticate(username=username, password=raw_password)
            user.managed_wallet = False

            if Group.objects.filter(name='Admin').exists():
                user.groups.add(Group.objects.get(name='Admin'))
            user.save()

            # create and provision org, including org wallet
            org_name = form.cleaned_data.get('org_name')
            org_role_name = form.cleaned_data.get('org_role_name')
            org_ico_url = form.cleaned_data.get('ico_url')
            org_role, created = IndyOrgRole.objects.get_or_create(name=org_role_name)
            org = org_signup(user, raw_password, org_name, org_role=org_role, org_ico_url=org_ico_url)

            # TODO need to auto-login with Atria custom user
            #login(request, user)

            return redirect('login')
    else:
        form = OrganizationSignUpForm()
    return render(request, 'registration/signup.html', {'form': form})


###############################################################
# UI views to support Django wallet login/logoff
###############################################################
def wallet_for_current_session(request):
    """
    Determine the current active wallet
    """

    wallet_name = request.session['wallet_name']
    wallet = IndyWallet.objects.filter(wallet_name=wallet_name).first()

    # validate it is the correct wallet
    wallet_type = request.session['wallet_type']
    wallet_owner = request.session['wallet_owner']
    if wallet_type == 'user':
        # verify current user owns wallet
        if wallet_owner == request.user.email:
            return wallet
        raise Exception('Error wallet/session config is not valid')
    elif wallet_type == 'org':
        # verify current user has relationship to org that owns wallet
        for org in request.user.indyrelationship_set.all():
            if org.org.org_name == wallet_owner:
                return wallet
        raise Exception('Error wallet/session config is not valid')
    else:
        raise Exception('Error wallet/session config is not valid')


###############################################################
# UI views to support wallet and agent UI functions
###############################################################
def profile_view(
    request,
    template=''
    ):
    """
    Example of user-defined view for Profile tab.
    """
    return render(request, 'indy/profile.html')

def data_view(
    request,
    template=''
    ):
    """
    Example of user-defined view for Data tab.
    """
    return render(request, 'indy/data.html')

def wallet_view(
    request,
    template=''
    ):
    """
    Example of user-defined view for Wallet tab.
    """
    return render(request, 'indy/wallet.html')


import importlib

def plugin_view(request, view_name):
    """
    Find and invoke user-defined view.
    These are configured in settings file.
    """

    view_function = getattr(settings, view_name)
    print(view_function)

    mod_name, func_name = view_function.rsplit('.',1)
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)

    return func(request)


######################################################################
# views to create and confirm agent-to-agent connections
######################################################################
def list_connections(
    request,
    template='indy/connection/list.html'
    ):
    """
    List Connections for the current wallet.
    """

    # expects a wallet to be opened in the current session
    wallet = wallet_for_current_session(request)
    connections = AgentConnection.objects.filter(wallet=wallet).all()
    return render(request, template, {'wallet_name': wallet.wallet_name, 'connections': connections})


def handle_connection_request(
    request,
    form_template='indy/connection/request.html',
    response_template='indy/connection/form_connection_info.html'
    ):
    """
    Send a Connection request (i.e. an Invitation).
    """

    if request.method=='POST':
        form = SendConnectionInvitationForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            partner_name = cd.get('partner_name')

            # get user or org associated with this wallet
            wallet = wallet_for_current_session(request)
            wallet_owner = request.session['wallet_owner']

            # get user or org associated with target partner
            target_user = get_user_model().objects.filter(email=partner_name).all()
            target_org = IndyOrganization.objects.filter(org_name=partner_name).all()

            if 0 < len(target_user):
                their_wallet = target_user[0].wallet
            elif 0 < len(target_org):
                their_wallet = target_org[0].wallet
            else:
                their_wallet = None

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            # build the connection and get the invitation data back
            try:
                my_connection = send_connection_invitation(wallet, partner_name)

                if their_wallet is not None:
                    their_connection = AgentConnection(
                        wallet = their_wallet,
                        partner_name = wallet_owner,
                        invitation = my_connection.invitation,
                        token = my_connection.token,
                        connection_type = 'Inbound',
                        status = 'Pending')
                    their_connection.save()

                if my_connection.wallet.wallet_org.get():
                    source_name = my_connection.wallet.wallet_org.get().org_name
                else:
                    source_name = my_connection.wallet.wallet_user.get().email
                target_name = my_connection.partner_name
                institution_logo_url = 'https://anon-solutions.ca/favicon.ico'
                return render(request, response_template, {'msg': 'Created invitation for ' + target_name, 'msg_txt': my_connection.invitation, 'msg_txt2': my_connection.token, 'msg_txt3': my_connection.invitation_shortform(source_name, target_name, institution_logo_url) })
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to create request for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to create invitation for ' + wallet.wallet_name})

    else:
        wallet = wallet_for_current_session(request)
        form = SendConnectionInvitationForm(initial={'wallet_name': wallet.wallet_name})

        return render(request, form_template, {'form': form})
    

def handle_connection_response(
    request,
    form_template='indy/connection/response.html',
    response_template='indy/form_response.html'
    ):
    """
    Respond to (Accept) a Connection request.
    """

    if request.method=='POST':
        form = SendConnectionResponseForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')
            partner_name = cd.get('partner_name')
            invitation_details = cd.get('invitation_details')

            # get user or org associated with this wallet
            wallet = wallet_for_current_session(request)

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            # build the connection and get the invitation data back
            try:
                my_connection = send_connection_confirmation(wallet, connection_id, partner_name, invitation_details)

                return render(request, response_template, {'msg': 'Updated connection for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update request for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update request for ' + wallet.wallet_name})

    else:
        # find connection request
        wallet = wallet_for_current_session(request)
        connection_id = request.GET.get('id', None)
        connections = []
        if connection_id:
            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
        if len(connections) > 0:
            form = SendConnectionResponseForm(initial={ 'connection_id': connection_id,
                                                        'wallet_name': connections[0].wallet.wallet_name, 
                                                        'partner_name': connections[0].partner_name, 
                                                        'invitation_details': connections[0].invitation })
        else:
            wallet = wallet_for_current_session(request)
            form = SendConnectionResponseForm(initial={'connection_id': 0, 'wallet_name': wallet.wallet_name})

        return render(request, form_template, {'form': form})
    

def poll_connection_status(
    request,
    form_template='indy/connection/status.html',
    response_template='indy/form_response.html'
    ):
    """
    Poll Connection status (normally a background task).
    """

    if request.method=='POST':
        form = PollConnectionStatusForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')

            # log out of current wallet, if any
            wallet = wallet_for_current_session(request)

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
            # TODO validate connection id
            my_connection = connections[0]

            # validate connection and get the updated status
            try:
                my_connection = check_connection_status(wallet, my_connection)

                return render(request, response_template, {'msg': 'Updated connection for ' + wallet.wallet_name + ', ' + my_connection.partner_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update request for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update request for ' + wallet.wallet_name})

    else:
        # find connection request
        wallet = wallet_for_current_session(request)
        connection_id = request.GET.get('id', None)
        connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()

        form = PollConnectionStatusForm(initial={ 'connection_id': connection_id,
                                                  'wallet_name': connections[0].wallet.wallet_name })

        return render(request, form_template, {'form': form})


def connection_qr_code(
    request, 
    token
    ):
    """
    Display a QR code for the given invitation.
    """

    # find connection for requested token
    connections = AgentConnection.objects.filter(token=token, connection_type='Outbound').all()
    if 0 == len(connections):
        return render(request, 'indy/form_response.html', {'msg': 'No connection found'})

    connection = connections[0]
    #qr = qrcode.QRCode(version=27, box_size=4)
    #qr.add_data(connection.invitation_shortform())
    #qr.make(fit=True)
    #image = qr.make_image()
    source_name = connection.partner_name
    target_name = connection.partner_name
    if connection.wallet.wallet_org.get():
        source_name = connection.wallet.wallet_org.get().org_name
        institution_logo_url = connection.wallet.wallet_org.get().ico_url
    else:
        source_name = connection.wallet.wallet_user.get().email
        institution_logo_url = None
    if not institution_logo_url:
        institution_logo_url = 'http://robohash.org/456'
    qr = pyqrcode.create(connection.invitation_shortform(source_name, target_name, institution_logo_url))
    path_to_image = '/tmp/'+token+'qr-offer.png'
    qr.png(path_to_image, scale=2, module_color=[0, 0, 0, 128], background=[0xff, 0xff, 0xff])
    image_data = open(path_to_image, "rb").read()

    # serialize to HTTP response
    response = HttpResponse(image_data, content_type="image/png")
    #image.save(response, "PNG")
    return response


######################################################################
# views to offer, request, send and receive credentials
######################################################################
def check_connection_messages(
    request,
    form_template='indy/connection/check_messages.html',
    response_template='indy/form_response.html'
    ):
    """
    Poll Connections for outstanding messages (normally a background task).
    """

    if request.method=='POST':
        form = PollConnectionStatusForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')

            # log out of current wallet, if any
            wallet = wallet_for_current_session(request)
    
            if connection_id > 0:
                connections = AgentConnection.objects.filter(wallet=wallet, id=connection_id).all()
            else:
                connections = AgentConnection.objects.filter(wallet=wallet).all()

            total_count = 0
            for connection in connections:
                # check for outstanding, un-received messages - add to outstanding conversations
                if connection.connection_type == 'Inbound':
                    msg_count = handle_inbound_messages(wallet, connection)
                    total_count = total_count + msg_count

            return render(request, response_template, {'msg': 'Received message count = ' + str(total_count)})

    else:
        # find connection request
        connection_id = request.GET.get('connection_id', None)
        wallet = wallet_for_current_session(request)
        if connection_id:
            connections = AgentConnection.objects.filter(wallet=wallet, id=connection_id).all()
        else:
            connection_id = 0
            connections = AgentConnection.objects.filter(wallet=wallet).all()
        # TODO validate connection id
        form = PollConnectionStatusForm(initial={ 'connection_id': connection_id,
                                                  'wallet_name': connections[0].wallet.wallet_name })

        return render(request, form_template, {'form': form})


def list_conversations(
    request,
    template='indy/conversation/list.html'
    ):
    """
    List Conversations for the current wallet.
    """

    # expects a wallet to be opened in the current session
    wallet = wallet_for_current_session(request)
    conversations = AgentConversation.objects.filter(connection__wallet=wallet).all()
    return render(request, template, {'wallet_name': wallet.wallet_name, 'conversations': conversations})


def handle_select_credential_offer(
    request,
    form_template='indy/credential/select_offer.html',
    response_template='indy/credential/offer.html'
    ):
    """
    Select a Credential Definition and display a form to enter Credential Offer information.
    """

    if request.method=='POST':
        form = SelectCredentialOfferForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')
            cred_def = cd.get('cred_def')
            partner_name = cd.get('partner_name')

            credential_name = cred_def.creddef_name
            credential_tag = cred_def.creddef_name

            # log out of current wallet, if any
            wallet = wallet_for_current_session(request)

            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
            # TODO validate connection id
            schema_attrs = cred_def.creddef_template
            form = SendCredentialOfferForm(initial={ 'connection_id': connection_id,
                                                     'wallet_name': connections[0].wallet.wallet_name,
                                                     'partner_name': partner_name,
                                                     'cred_def': cred_def.id,
                                                     'schema_attrs': schema_attrs,
                                                     'credential_name': credential_name,
                                                     'credential_tag': credential_tag })

            return render(request, response_template, {'form': form})

    else:
        # find conversation request
        connection_id = request.GET.get('connection_id', None)
        wallet = wallet_for_current_session(request)
        connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
        # TODO validate connection id
        form = SelectCredentialOfferForm(initial={ 'connection_id': connection_id,
                                                   'partner_name': connections[0].partner_name,
                                                   'wallet_name': connections[0].wallet.wallet_name})

        return render(request, form_template, {'form': form})


def handle_credential_offer(
    request,
    template='indy/form_response.html'
    ):
    """
    Send a Credential Offer.
    """

    if request.method=='POST':
        form = SendCredentialOfferForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')
            cred_def_id = cd.get('cred_def')
            credential_name = cd.get('credential_name')
            credential_tag = cd.get('credential_tag')
            schema_attrs = cd.get('schema_attrs')
            schema_attrs = json.loads(schema_attrs)
            for attr in schema_attrs:
                field_name = 'schema_attr_' + attr
                field_value = request.POST.get(field_name)
                schema_attrs[attr] = field_value
            schema_attrs = json.dumps(schema_attrs)

            wallet = wallet_for_current_session(request)
    
            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
            # TODO validate connection id
            my_connection = connections[0]

            cred_defs = IndyCredentialDefinition.objects.filter(id=cred_def_id, wallet=wallet).all()
            cred_def = cred_defs[0]

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            # build the credential offer and send
            try:
                my_conversation = send_credential_offer(wallet, my_connection, credential_tag, json.loads(schema_attrs), cred_def, credential_name)

                return render(request, template, {'msg': 'Updated conversation for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update conversation for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update conversation for ' + wallet.wallet_name})

    else:
        return render(request, 'indy/form_response.html', {'msg': 'Method not allowed'})


def handle_cred_offer_response(
    request,
    form_template='indy/credential/offer_response.html',
    response_template='indy/form_response.html'
    ):
    """
    Respond to a Credential Offer by sending a Credential Request.
    """

    if request.method=='POST':
        form = SendCredentialResponseForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            conversation_id = cd.get('conversation_id')

            wallet = wallet_for_current_session(request)
    
            # find conversation request
            conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
            my_conversation = conversations[0]
            # TODO validate conversation id
            my_connection = my_conversation.connection

            # build the credential request and send
            try:
                my_conversation = send_credential_request(wallet, my_connection, my_conversation)

                return render(request, response_template, {'msg': 'Updated conversation for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update conversation for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update conversation for ' + wallet.wallet_name})

    else:
        # find conversation request, fill in form details
        conversation_id = request.GET.get('conversation_id', None)
        wallet = wallet_for_current_session(request)
        conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
        # TODO validate conversation id
        conversation = conversations[0]
        indy_conversation = json.loads(conversation.conversation_data)
        # TODO validate connection id
        connection = conversation.connection
        form = SendCredentialResponseForm(initial={ 
                                                 'conversation_id': conversation_id,
                                                 'wallet_name': connection.wallet.wallet_name,
                                                 'from_partner_name': connection.partner_name,
                                                 'claim_id':indy_conversation['claim_id'],
                                                 'claim_name': indy_conversation['claim_name'],
                                                 'credential_attrs': indy_conversation['credential_attrs'],
                                                 'libindy_offer_schema_id': json.loads(indy_conversation['libindy_offer'])['schema_id']
                                                })

        return render(request, form_template, {'form': form})


######################################################################
# views to request, send and receive proofs
######################################################################
def handle_proof_req_response(
    request,
    form_template='indy/proof/send_response.html',
    response_template='indy/proof/select_claims.html'
    ):
    """
    First stage in responding to a Proof Request - confirm to search for claims.
    """

    if request.method=='POST':
        form = SendProofReqResponseForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            conversation_id = cd.get('conversation_id')
            proof_req_name = cd.get('proof_req_name')

            wallet = wallet_for_current_session(request)
    
            # find conversation request
            conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
            my_conversation = conversations[0]
            # TODO validate conversation id
            # TODO validate connection id
            my_connection = my_conversation.connection

            # find claims for this proof request and display for the user
            try:
                claim_data = get_claims_for_proof_request(wallet, my_connection, my_conversation)

                form = SelectProofReqClaimsForm(initial={
                         'conversation_id': conversation_id,
                         'wallet_name': my_connection.wallet.wallet_name,
                         'from_partner_name': my_connection.partner_name,
                         'proof_req_name': proof_req_name,
                         'requested_attrs': claim_data,
                    })

                return render(request, response_template, {'form': form})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to find claims for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to find claims for ' + wallet.wallet_name})

    else:
        # find conversation request, fill in form details
        wallet = wallet_for_current_session(request)
        conversation_id = request.GET.get('conversation_id', None)
        conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
        # TODO validate conversation id
        conversation = conversations[0]
        indy_conversation = json.loads(conversation.conversation_data)
        # TODO validate connection id
        connection = conversation.connection
        form = SendProofReqResponseForm(initial={ 
                                                 'conversation_id': conversation_id,
                                                 'wallet_name': connection.wallet.wallet_name,
                                                 'from_partner_name': connection.partner_name,
                                                 'proof_req_name': indy_conversation['proof_request_data']['name'],
                                                })

    return render(request, form_template, {'form': form})


def handle_proof_select_claims(
    request,
    template='indy/form_response.html'
    ):
    """
    Select claims to construct Proof for Proof Request.
    """

    if request.method=='POST':
        form = SelectProofReqClaimsForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            conversation_id = cd.get('conversation_id')
            proof_req_name = cd.get('proof_req_name')

            wallet = wallet_for_current_session(request)

            # find conversation request
            conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
            # TODO validate conversation id
            my_conversation = conversations[0]
            indy_conversation = json.loads(my_conversation.conversation_data)
            # TODO validate connection id
            my_connection = my_conversation.connection

            # get selected attributes for proof request
            requested_attributes = indy_conversation['proof_request_data']['requested_attributes']
            requested_predicates = indy_conversation['proof_request_data']['requested_predicates']
            credential_attrs = {}
            for attr in requested_attributes:
                field_name = 'proof_req_attr_' + attr
                value = request.POST.get(field_name)
                if value.startswith('ref::'):
                    credential_attrs[attr] = {'referent': value.replace('ref::','')}
                else:
                    credential_attrs[attr] = {'value': value}
            for attr in requested_predicates:
                field_name = 'proof_req_attr_' + attr
                value = request.POST.get(field_name)
                if value.startswith('ref::'):
                    credential_attrs[attr] = {'referent': value.replace('ref::','')}
                else:
                    credential_attrs[attr] = {'value': value}

            # send claims for this proof request to requestor
            try:
                proof_data = send_claims_for_proof_request(wallet, my_connection, my_conversation, credential_attrs)

                return render(request, template, {'msg': 'Sent proof request for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to find claims for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to find claims for ' + wallet.wallet_name})

    else:
        return render(request, 'indy/form_response.html', {'msg': 'Method not allowed'})


def poll_conversation_status(
    request,
    form_template='indy/conversation/status.html',
    response_template='indy/form_response.html'
    ):
    """
    Poll Conversation status (normally a background task).
    """

    if request.method=='POST':
        form = PollConversationStatusForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            conversation_id = cd.get('conversation_id')

            wallet = wallet_for_current_session(request)
    
            # find conversation request
            conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
            # TODO validate conversation id
            my_conversation = conversations[0]
            indy_conversation = json.loads(my_conversation.conversation_data)
            # TODO validate connection id
            my_connection = my_conversation.connection

            # check conversation status
            try:
                polled_count = poll_message_conversation(wallet, my_connection, my_conversation)

                return render(request, response_template, {'msg': 'Updated conversation for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update conversation for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update conversation for ' + wallet.wallet_name})

    else:
        # find conversation request, fill in form details
        wallet = wallet_for_current_session(request)
        conversation_id = request.GET.get('conversation_id', None)
        conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
        # TODO validate conversation id
        conversation = conversations[0]
        indy_conversation = json.loads(conversation.conversation_data)
        # TODO validate connection id
        connection = conversation.connection
        form = PollConversationStatusForm(initial={'conversation_id': conversation_id, 'wallet_name': connection.wallet.wallet_name})

    return render(request, form_template, {'form': form})


def handle_select_proof_request(
    request,
    form_template='indy/proof/select_request.html',
    response_template='indy/proof/send_request.html'
    ):
    """
    Select a Proof Request to send, based on the templates available in the database.
    """

    if request.method=='POST':
        form = SelectProofRequestForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            proof_request = cd.get('proof_request')
            connection_id = cd.get('connection_id')
            partner_name = cd.get('partner_name')

            wallet = wallet_for_current_session(request)

            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
            connection = connections[0]
            connection_data = json.loads(connection.connection_data)
            institution_did = connection_data['data']['public_did']

            proof_req_attrs = proof_request.proof_req_attrs
            proof_req_predicates = proof_request.proof_req_predicates

            # selective attribute substitutions
            proof_req_attrs = proof_req_attrs.replace('$ISSUER_DID', institution_did)
            proof_req_predicates = proof_req_predicates.replace('$ISSUER_DID', institution_did)

            proof_form = SendProofRequestForm(initial={
                    'wallet_name': connection.wallet.wallet_name,
                    'connection_id': connection_id,
                    'partner_name': partner_name,
                    'proof_name': proof_request.proof_req_name,
                    'proof_attrs': proof_req_attrs,
                    'proof_predicates': proof_req_predicates})

            return render(request, response_template, {'form': proof_form})

    else:
        # find conversation request
        wallet = wallet_for_current_session(request)
        connection_id = request.GET.get('connection_id', None)
        connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
        connection = connections[0]
        form = SelectProofRequestForm(initial={ 'connection_id': connection_id,
                                                'partner_name': connection.partner_name,
                                                'wallet_name': connection.wallet.wallet_name })

        return render(request, form_template, {'form': form})


def handle_send_proof_request(
    request,
    template='indy/form_response.html'
    ):
    """
    Send a Proof Request for the selected Proof Request.
    User can edit the requested attributes and predicates.
    """

    if request.method=='POST':
        form = SendProofRequestForm(request.POST)
        if not form.is_valid():
            return render(request, 'indy/form_response.html', {'msg': 'Form error', 'msg_txt': str(form.errors)})
        else:
            cd = form.cleaned_data
            connection_id = cd.get('connection_id')
            proof_name = cd.get('proof_name')
            proof_attrs = cd.get('proof_attrs')
            proof_predicates = cd.get('proof_predicates')
            proof_uuid = str(uuid.uuid4())

            wallet = wallet_for_current_session(request)
    
            connections = AgentConnection.objects.filter(id=connection_id, wallet=wallet).all()
            # TODO validate connection id
            my_connection = connections[0]

            # set wallet password
            # TODO vcx_config['something'] = raw_password

            # build the proof request and send
            try:
                my_conversation = send_proof_request(wallet, my_connection, proof_uuid, proof_name, json.loads(proof_attrs), json.loads(proof_predicates))

                return render(request, template, {'msg': 'Updated conversation for ' + wallet.wallet_name})
            except IndyError:
                # ignore errors for now
                print(" >>> Failed to update conversation for", wallet.wallet_name)
                return render(request, 'indy/form_response.html', {'msg': 'Failed to update conversation for ' + wallet.wallet_name})

    else:
        return render(request, 'indy/form_response.html', {'msg': 'Method not allowed'})


def handle_view_proof(
    request,
    template='indy/proof/view_proof.html'
    ):
    """
    View the Proof sent by the Prover.
    """

    wallet = wallet_for_current_session(request)
    conversation_id = request.GET.get('conversation_id', None)
    conversations = AgentConversation.objects.filter(id=conversation_id, connection__wallet=wallet).all()
    # TODO validate conversation id
    conversation = conversations[0]
    return render(request, template, {'conversation': json.loads(conversation.conversation_data)})


######################################################################
# views to list wallet credentials
######################################################################
def form_response(request):
    """
    Generic response page.
    """

    msg = request.GET.get('msg', None)
    msg_txt = request.GET.get('msg_txt', None)
    return render(request, 'indy/form_response.html', {'msg': msg, 'msg_txt': msg_txt})


def list_wallet_credentials(
    request
    ):
    """
    List all credentials in the current wallet.
    """

    wallet_handle = None
    try:
        wallet = wallet_for_current_session(request)
        raw_password = request.session['wallet_password']
        wallet_handle = open_wallet(wallet.wallet_name, raw_password)

        (search_handle, search_count) = run_coroutine_with_args(prover_search_credentials, wallet_handle, "{}")
        credentials = run_coroutine_with_args(prover_fetch_credentials, search_handle, search_count)
        run_coroutine_with_args(prover_close_credentials_search, search_handle)

        return render(request, 'indy/credential/list.html', {'wallet_name': wallet.wallet_name, 'credentials': json.loads(credentials)})
    except:
        raise
    finally:
        if wallet_handle:
            close_wallet(wallet_handle)




from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

import asyncio
import random
import string


######################################################################
# Indy model override utilities
######################################################################

def get_indy_settings_model(model_type):
    """
    Return the Indy override model that is active in this project.
    model_type is 'INDY_ORGANIZATION_MODEL' or 'INDY_ORG_RELATION_MODEL'
    """

    try:
        return django_apps.get_model(getattr(settings, model_type), require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("%s must be of the form 'app_label.model_name'" % model_type)
    except LookupError:
        raise ImproperlyConfigured(
            "%s refers to model '%s' that has not been installed" % (model_type, getattr(settings, model_type))
        )

def get_indy_model(model_type):
    """
    Return the Indy override model that is active in this project.
    model_type is 'INDY_ORGANIZATION_MODEL' or 'INDY_ORG_RELATION_MODEL'
    """
    
    try:
        return django_apps.get_model(model_type, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("%s must be of the form 'app_label.model_name'" % model_type)
    except LookupError:
        raise ImproperlyConfigured(
            "refers to model '%s' that has not been installed" % model_type
        )


######################################################################
# a few random utilities
######################################################################

def random_int(low, high):
    return random.randint(low, high)

def random_alpha_string(length, contains_spaces=False):
    if contains_spaces:
        chars = string.ascii_uppercase + ' '
    else:
        chars = string.ascii_uppercase
    return ''.join(random.SystemRandom().choice(chars) for _ in range(length))

def random_numeric_string(length):
    chars = string.digits
    return ''.join(random.SystemRandom().choice(chars) for _ in range(length))

def random_an_string(length, contains_spaces=False):
    if contains_spaces:
        chars = string.ascii_uppercase + string.digits + ' '
    else:
        chars = string.ascii_uppercase + string.digits
    return ''.join(random.SystemRandom().choice(chars) for _ in range(length))

def random_schema_version():
    version = format("%d.%d.%d" % (random.randint(1, 101), random.randint(1, 101), random.randint(1, 101)))
    return version


######################################################################
# coroutine utilities
######################################################################

def run_coroutine(coroutine):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coroutine())
    finally:
        loop.close()

def run_coroutine_with_args(coroutine, *args):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coroutine(*args))
    finally:
        loop.close()

def run_coroutine_with_kwargs(coroutine, *args, **kwargs):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coroutine(*args, **kwargs))
    finally:
        loop.close()




from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.views.generic.edit import UpdateView

from ..wallet_utils import *
from ..indy_utils import *


class IndyDIDTests(TestCase):
    """
    Tests for creating ledger DID's
    """

    def test_wallet_did_create(self):
        # create a random wallet name
        user_name = random_alpha_string(10) + "@" + random_alpha_string(10) + ".com"
        user_wallet_name = get_user_wallet_name(user_name)
        nym_info = create_and_register_did(user_wallet_name, "NotTrustee")




from django.core.management.base import BaseCommand
from django.conf import settings
from django.utils import timezone
from django.contrib.auth import get_user_model

import yaml
import os

from indy_community.utils import *
from indy_community.models import *
from indy_community.agent_utils import *
from indy_community.registration_utils import *


USER_ROLE = getattr(settings, "DEFAULT_USER_ROLE", 'User')
ORG_ROLE = getattr(settings, "DEFAULT_ORG_ROLE", 'Admin')

def get_attr_value(key):
    if isinstance(key, dict):
        object_type = None
        object_attrs = {}
        for attr_attr in key:
            if attr_attr == 'class':
                object_type = key['class']
            else:
                object_attrs[attr_attr] = key[attr_attr]
        return get_indy_model(object_type).objects.filter(**object_attrs).get()
    else:
        return key

class Command(BaseCommand):
    help = 'Loads Organizations and optionally creates Credential Definitions'

    def add_arguments(self, parser):
        parser.add_argument('config_file', nargs='+')

    def handle(self, *args, **options):
        schemas = None
        org = None

        # verify that config file exists and we can load yaml
        self.stdout.write("config_file = %s" % str(options['config_file']))
        with open(str(options['config_file'][0]), 'r') as stream:
            try:
                orgs = yaml.load(stream)
            except yaml.YAMLError as exc:
                self.stdout.write(exc)
                raise

            # validate data in yaml file
            for name in orgs:
                org = orgs[name]
                # TODO validation

        # now create orgs (and potentially create cred defs)
        if orgs:
            for name in orgs:
                org = orgs[name]
                first_name = org['first_name']
                last_name = org['last_name']
                email = org['email']
                password = org['password']
                role_name = org['role']
                if 'ico_url' in org:
                    ico_url = org['ico_url']
                else:
                    ico_url = None

                if "$random" in name:
                    name = name.replace("$random", random_alpha_string(12))
                if "$random" in email:
                    email = email.replace("$random", random_alpha_string(12))

                user_attrs = {}
                if 'user' in org:
                    for attr in org['user']:
                        user_attrs[attr] = get_attr_value(org['user'][attr])
                user = get_user_model().objects.create_user(first_name=first_name, last_name=last_name, email=email, password=password, **user_attrs)
                user.groups.add(Group.objects.get(name=ORG_ROLE))
                user.save()

                org_attrs = {}
                if 'org' in org:
                    for attr in org['org']:
                        org_attrs[attr] = get_attr_value(org['org'][attr])
                relation_attrs = {}
                if 'relation' in org:
                    for attr in org['relation']:
                        relation_attrs[attr] = get_attr_value(org['relation'][attr])
                org_role, created = IndyOrgRole.objects.get_or_create(name=role_name)
                org = org_signup(user, password, name, org_attrs=org_attrs, org_relation_attrs=relation_attrs, org_role=org_role, org_ico_url=ico_url)
